#!/usr/bin/env python
# coding=utf-8

from modules.core.classes import ExperimentalDrop, ExperimentalSetup
from utils.enums import FittingMethod

# from utils.keymap import *

from typing import List, Tuple
import cv2
import numpy as np
import matplotlib.pyplot as plt
import tkinter.messagebox as msgbox
import tkinter.simpledialog as simpledialog

# from __future__ import print_function


# from subprocess import call
# import numpy as np
# import time
# import datetime
# from Tkinter import *
# import tkFileDialog

import sys
from scipy import optimize  # DS 7/6/21 - for least squares fit
import tensorflow as tf  # DS 9/6/21 - for loading ML model

from modules.preprocessing.preprocessing import prepare_hydrophobic, tilt_correction
from utils.config import *
from utils.geometry import Rect2

# import os

MAX_IMAGE_TO_SCREEN_RATIO = 0.8


def get_ift_regions(
    img: np.ndarray,
    padding: int = 5,
    area_thresh: int = 1,
    drop_frac: float = 1,
    scharr_block: int = 5,
    canny1: float = 80,
    canny2: float = 160,
) -> Tuple[np.ndarray, tuple]:
    """
    Automatically detects the drop and needle regions in an image.
    """

    original = img.copy()
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    H, W = gray.shape
    # Applying 7x7 Gaussian Blur
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    # 1) rough mask via inverted Otsu + largest CC
    threshold = cv2.threshold(
        gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]
    analysis = cv2.connectedComponentsWithStats(threshold, 8)
    (totalLabels, label_ids, stats, centroids) = analysis
    if totalLabels <= 1:
        print("drops found.")
        return original, None

    # Filter for components entering from the top and find the largest among them
    top_entering_blobs_indices = []
    for i in range(1, totalLabels):  # Iterate through all components (0 is background)
        if stats[i, cv2.CC_STAT_TOP] == 0:
            top_entering_blobs_indices.append(i)

    if not top_entering_blobs_indices:
        print("No blobs found entering from the top of the frame.")
        return original, None

    # Select the largest blob among those entering from the top
    blob = -1
    max_area_found = -1
    for idx in top_entering_blobs_indices:
        area = stats[idx, cv2.CC_STAT_AREA]
        if area > max_area_found:
            max_area_found = area
            blob = idx

    if blob == -1:  # Should not happen if top_entering_blobs_indices was populated
        print("Error selecting largest top-entering blob.")
        return original, None

    if stats[blob, cv2.CC_STAT_AREA] < area_thresh:
        print(
            f"Largest top-entering blob (ID: {blob}, Area: {stats[blob, cv2.CC_STAT_AREA]}) too small (threshold: {area_thresh})."
        )
        return original, None

    x, y, bw, bh, _ = stats[blob]
    # print(f"Blob ID: {blob}, Area: {stats[blob, cv2.CC_STAT_AREA]}, Bounding box: (x= {x}, y= {y}) - (w= {bw}, h= {bh})")
    roi = gray[y: y + bh, x: x + bw]
    roi_mask = (label_ids[y: y + bh, x: x + bw] == blob).astype(np.uint8) * 255

    # 2) Scharr→adaptiveThresh→Canny inside ROI
    blur = cv2.GaussianBlur(roi, (scharr_block, scharr_block), 0)
    dx = cv2.Scharr(blur, cv2.CV_16S, 1, 0)
    dy = cv2.Scharr(blur, cv2.CV_16S, 0, 1)
    grad = dx.astype(float) ** 2 + dy.astype(float) ** 2
    grad = np.uint8(255 * (grad / grad.max()))
    cv2.adaptiveThreshold(
        grad,
        255,
        cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        cv2.THRESH_BINARY,
        blockSize=scharr_block,
        C=0,
        dst=grad,
    )
    grad = cv2.bitwise_and(grad, grad, mask=roi_mask)
    edges = cv2.Canny(grad, canny1, canny2)

    # 3) collect edge points *only* in the bottom (1-drop_frac) of ROI:
    pts = np.column_stack(np.nonzero(edges))  # (row, col)
    if len(pts) < 20:
        print("Too few edges.")
        return original, None
    # Calculate width of the blob at each row within the ROI
    blob_widths_at_each_row = np.sum(roi_mask // 255, axis=1)
    # Define the upper part of the blob to estimate needle width
    # Consider top 15% of blob height, but at least 3px and at most 20px
    initial_needle_scan_height = max(3, min(int(bh * 0.15), 500))
    # Ensure scan height is less than blob height for sensible indexing
    if initial_needle_scan_height >= bh and bh > 0:
        initial_needle_scan_height = bh - 1
    elif bh == 0:  # Should not happen if blob found
        initial_needle_scan_height = 0
    avg_needle_width = 0
    if initial_needle_scan_height > 0:
        candidate_needle_widths = blob_widths_at_each_row[:initial_needle_scan_height]
        if np.any(candidate_needle_widths > 0):
            avg_needle_width = np.median(
                candidate_needle_widths[candidate_needle_widths > 0]
            )

    if (
        avg_needle_width <= 1
    ):  # If needle width is too small or not found, use a small portion of top widths
        fallback_scan_height = max(1, int(bh * 0.05))
        if bh > 0 and np.any(blob_widths_at_each_row[:fallback_scan_height] > 0):
            avg_needle_width = np.mean(
                blob_widths_at_each_row[:fallback_scan_height][
                    blob_widths_at_each_row[:fallback_scan_height] > 0
                ]
            )
        if avg_needle_width <= 0:  # Absolute fallback if still no valid width
            avg_needle_width = 1
    expansion_factor = 1  # Drop should be significantly wider than needle
    search_start_row = initial_needle_scan_height

    found_expansion_point = False
    calculated_row_thresh = 0
    for current_row in range(search_start_row, bh):
        # Add small absolute diff
        if (
            blob_widths_at_each_row[current_row] > expansion_factor *
                avg_needle_width
            and blob_widths_at_each_row[current_row] > avg_needle_width + 2
        ):
            calculated_row_thresh = current_row
            found_expansion_point = True
            # print(f"Dynamic row_thresh: Initial needle width ~{avg_needle_width:.1f}px. Expansion at row {current_row} (width {blob_widths_at_each_row[current_row]}px).")
            break

    if not found_expansion_point:
        print(
            f"No clear expansion point found. Initial needle width ~{avg_needle_width:.1f}px. Using drop_frac based fallback: {calculated_row_thresh}."
        )
        pass  # Already defaulted to drop_frac based
    row_thresh = calculated_row_thresh

    # print(f"Row threshold for drop detection: {row_thresh} (calculated)")
    # --- End of dynamic row_thresh determination ---
    def create_bounds_array(x1, y1, x2, y2):
        # Ensure coordinates are integers for array creation if they are not None
        if any(v is None for v in [x1, y1, x2, y2]):
            return None
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        return np.array(
            [
                [[x1, y1, x2, y1]],  # top edge
                [[x2, y1, x2, y2]],  # right edge
                [[x2, y2, x1, y2]],  # bottom edge
                [[x1, y2, x1, y1]],  # left edge
            ],
            dtype=np.int32,
        )

    # Create the bounding box for the entire ROI
    x1_roi = x
    y1_roi = y
    x2_roi = x + bw
    y2_roi = y + bh
    # Create the absolute bounding box for the entire ROI
    abs_bounds = create_bounds_array(x1_roi, y1_roi, x2_roi, y2_roi)
    # Create the cropped bounding box for the entire ROI
    crop_bounds = create_bounds_array(x1_roi, y1_roi, x2_roi, y2_roi)
    crop_roi = original[y1_roi:y2_roi, x1_roi:x2_roi]
    if not found_expansion_point:
        needle_pts = pts[pts[:, 0] < row_thresh]
        pts_xy_needle = needle_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_needle.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the needle points, *relative to the ROI's top-left corner*
            needle_x1_roi = np.min(pts_xy_needle[:, 0])
            needle_x2_roi = np.max(pts_xy_needle[:, 0])
            needle_y1_roi = np.min(pts_xy_needle[:, 1])
            needle_y2_roi = np.max(pts_xy_needle[:, 1])
            needle_abs_bounds = create_bounds_array(
                x + needle_x1_roi - padding,
                y + needle_y1_roi,
                x + needle_x2_roi + padding,
                y + needle_y2_roi - padding,
            )
            needle_crop_bounds = create_bounds_array(
                needle_x1_roi - padding,
                needle_y1_roi,
                needle_x2_roi + padding,
                needle_y2_roi - padding,
            )
            # Crop the original image using the final absolute bounding box coordinates
            crop = original[
                needle_y1_roi:needle_y2_roi, x + needle_x1_roi: x + needle_x2_roi
            ]
            needle_rect = Rect2(
                (x + needle_x1_roi, y + needle_y1_roi),
                (x + needle_x2_roi, y + needle_y2_roi),
            )
            return needle_rect
    else:
        needle_pts = pts[pts[:, 0] < row_thresh]
        pts_xy_needle = needle_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_needle.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the needle points, *relative to the ROI's top-left corner*
            needle_x1_roi = np.min(pts_xy_needle[:, 0]) - padding
            needle_x2_roi = np.max(pts_xy_needle[:, 0]) + padding
            needle_y1_roi = np.min(pts_xy_needle[:, 1])
            needle_y2_roi = np.max(pts_xy_needle[:, 1]) - padding
            needle_abs_bounds = create_bounds_array(
                x + needle_x1_roi,
                y + needle_y1_roi,
                x + needle_x2_roi,
                y + needle_y2_roi,
            )
            needle_crop_bounds = create_bounds_array(
                needle_x1_roi - padding, needle_y1_roi, needle_x2_roi, needle_y2_roi
            )
            needle_rect = Rect2(
                (x + needle_x1_roi, y + needle_y1_roi),
                (x + needle_x2_roi, y + needle_y2_roi),
            )

        drop_pts = pts[pts[:, 0] >= row_thresh]
        pts_xy_drop = drop_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_drop.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the drop points, *relative to the ROI's top-left corner*
            drop_x1_roi = np.min(pts_xy_drop[:, 0]) - padding
            drop_x2_roi = np.max(pts_xy_drop[:, 0]) + padding
            drop_y1_roi = np.min(pts_xy_drop[:, 1])  # - padding
            drop_y2_roi = np.max(pts_xy_drop[:, 1]) + padding
            drop_abs_bounds = create_bounds_array(
                x + drop_x1_roi, y + drop_y1_roi, x + drop_x2_roi, y + drop_y2_roi
            )
            drop_crop_bounds = create_bounds_array(
                drop_x1_roi, drop_y1_roi, drop_x2_roi, drop_y2_roi
            )
            drop_rect = Rect2(
                (x + drop_x1_roi, y + drop_y1_roi), (x + drop_x2_roi, y + drop_y2_roi)
            )

        crop = original[drop_y1_roi:drop_y2_roi,
                        x + drop_x1_roi: x + drop_x2_roi]
        return drop_rect, needle_rect


def set_drop_region(
    experimental_drop: ExperimentalDrop,
    experimental_setup: ExperimentalSetup,
    index: int = 0,
) -> None:
    # select the drop and needle regions in the image
    screen_size = experimental_setup.screen_resolution
    image_size = experimental_drop.image.shape
    scale: float = set_scale(image_size, screen_size)
    screen_position = set_screen_position(screen_size)

    if experimental_setup.drop_id_method == RegionSelect.AUTOMATED:
        from modules.preprocessing.preprocessing import auto_crop

        experimental_drop.cropped_image, (left, right, top, bottom) = auto_crop(
            experimental_drop.image
        )
        # print("experimental_drop.cropped_image",experimental_drop.cropped_image is None)
        if experimental_setup.original_boole == 1:  # show found drop

            # Clear all existing figures to avoid conflicts with residual plots
            plt.close("all")
            fig = plt.figure()  # Explicitly create a new figure window
            plt.title(f"Original image {index}")
            plt.imshow(experimental_drop.image)
            plt.show()  # ✅ Block execution until the window is manually closed by the user
            plt.close(fig)  # clean up and close the figure after it's shown

        if experimental_setup.cropped_boole == 1:
            # Clear all existing figures to avoid conflicts with residual plots
            plt.close("all")
            fig = plt.figure()  # Explicitly create a new figure window
            plt.title(f"Cropped image {index}")
            plt.imshow(experimental_drop.cropped_image)
            plt.show()
            plt.close(fig)
        experimental_setup.drop_region = [(left, top), (right, bottom)]
    elif experimental_setup.drop_id_method == RegionSelect.USER_SELECTED:
        experimental_setup.drop_region, experimental_drop.cropped_image = (
            user_select_region(
                experimental_drop.image,
                f"Select drop region for Image {index}",
                scale,
                screen_position,
            )
        )


def find_image_edge(img: np.ndarray, low=50, high=150, apertureSize=3):
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    edges = cv2.Canny(gray, low, high, apertureSize=apertureSize)
    return edges


def user_select_region(
    image: np.ndarray, title: str, scale: float, screen_position
) -> Tuple[Rect2, np.ndarray]:
    """
    Function to manually select a region of interest (ROI) in an image using mouse events.
    User can click and drag to create a rectangle, and press Enter to confirm the selection.
    Returns:
        - A Rect2 object representing the selected region.
        - The cropped image based on the selected region.
    """
    region = user_roi(image, title, scale, screen_position)
    cropped_image = image_crop(image, region)
    return (region, cropped_image)


def set_needle_region(
    experimental_drop: ExperimentalDrop,
    experimental_setup: ExperimentalSetup,
    center_x: float = None,
    index: int = 0,
) -> None:
    screen_size = experimental_setup.screen_resolution
    image_size = experimental_drop.image.shape
    scale = set_scale(image_size, screen_size)
    screen_position = set_screen_position(screen_size)

    if experimental_setup.needle_region_method == RegionSelect.AUTOMATED:
        img = experimental_drop.image

        # find edges in the image
        edges = find_image_edge(img)

        height, width = img.shape[:2]
        cx = center_x if center_x is not None else width // 2

        lines = cv2.HoughLines(edges, 1, np.pi / 180, 100)
        vertical_lines = []

        # Collect near-vertical lines in upper-middle region
        if lines is not None:
            for line in lines:
                rho, theta = line[0]
                deg = np.rad2deg(theta)

                if deg < 5 or deg > 175:  # very vertical
                    a = np.cos(theta)
                    b = np.sin(theta)
                    x0 = a * rho
                    y0 = b * rho
                    x1 = int(x0 + 1000 * (-b))
                    y1 = int(y0 + 1000 * (a))
                    x2 = int(x0 - 1000 * (-b))
                    y2 = int(y0 - 1000 * (a))

                    y_min = min(y1, y2)
                    y_max = max(y1, y2)
                    x_avg = (x1 + x2) // 2

                    # Must be near center and top
                    if abs(x_avg - cx) < width * 0.4 and y_min < height * 0.2:
                        vertical_lines.append(
                            {
                                "x": x_avg,
                                "theta": theta,
                                "y_min": y_min,
                                "y_max": y_max,
                                "pts": ((x1, y1), (x2, y2)),
                            }
                        )

        # If exactly 2 vertical lines were found, just use them
        if len(vertical_lines) == 2:
            l1, l2 = vertical_lines[0], vertical_lines[1]
            best_pair = (l1, l2)
        else:
            # Try to find a good pair
            best_pair = None
            best_score = float("inf")

            for i in range(len(vertical_lines)):
                for j in range(i + 1, len(vertical_lines)):
                    l1 = vertical_lines[i]
                    l2 = vertical_lines[j]

                    # Height similarity
                    height_diff = abs(
                        (l1["y_max"] - l1["y_min"]) -
                        (l2["y_max"] - l2["y_min"])
                    )
                    if height_diff > 20:
                        continue

                    # Theta similarity
                    angle_diff = abs(l1["theta"] - l2["theta"])
                    if angle_diff > np.deg2rad(1):  # ~1 degree
                        continue

                    # Prefer taller + closer lines
                    avg_height = (
                        l1["y_max"] - l1["y_min"] + l2["y_max"] - l2["y_min"]
                    ) / 2
                    score = angle_diff + height_diff / 10 - avg_height / 50

                    if score < best_score:
                        best_score = score
                        best_pair = (l1, l2)

        if best_pair:
            x1 = min(best_pair[0]["x"], best_pair[1]["x"])
            x2 = max(best_pair[0]["x"], best_pair[1]["x"])
            y_top = min(best_pair[0]["y_min"], best_pair[1]["y_min"])
            y_bot = max(best_pair[0]["y_max"], best_pair[1]["y_max"])

            width = x2 - x1
            height = y_bot - y_top

            x_padding = int(0.2 * width)
            y_padding = int(0.2 * height)

            x1 = max(x1 - x_padding, 0)
            x2 = x2 + x_padding
            y_top = y_top + y_padding  # subtract padding from top = move it down
            # subtract padding from bottom = move it up
            y_bot = max(y_bot - y_padding, y_top)

            # Optional: draw
            # cv2.rectangle(img, (x1, y_top), (x2, y_bot), (0, 255, 0), 2)

            experimental_setup.needle_region = Rect2((x1, y_top), (x2, y_bot))
            return

        print("error: can't find needle. Please use user-selected method")

    elif experimental_setup.needle_region_method == RegionSelect.USER_SELECTED:
        experimental_setup.needle_region, experimental_drop.cropped_image = (
            user_select_region(
                experimental_drop.image,
                f"Select drop region for Image {index}",
                scale,
                screen_position,
            )
        )


def crop_needle(img: np.ndarray) -> np.ndarray:
    padding = 10
    angle_tolerance = 15
    original_img = img.copy()
    # adjustable parameters
    vertical_pct = 0.8  # fraction of height to keep from top
    horizontal_pct = 0.5  # fraction of width to keep, centered
    # initial crop of the image
    h, w = img.shape[:2]
    top = int(h * vertical_pct)
    side_margin = int((w * (1 - horizontal_pct)) / 2)
    img = img[:top, side_margin: w - side_margin]

    # a veriable used to eliminate the vertical lines that exist in the bottom half of the image
    bottom_verticals_threshold = 0.7
    # 1. Preprocess: grayscale & edges
    # if img.ndim == 3 else img.copy()
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    edges = cv2.Canny(blurred, 50, 180, apertureSize=3)

    # 2. Detect line segments via probabilistic Hough
    lines = cv2.HoughLinesP(
        edges, rho=1, theta=np.pi / 180, threshold=45, minLineLength=1, maxLineGap=50
    )
    if lines is None or len(lines) < 2:
        print("Insufficient line segments detected.")
        return original_img

    # 3. Optimized vertical filter
    dx = lines[:, 0, 2] - lines[:, 0, 0]
    dy = lines[:, 0, 3] - lines[:, 0, 1]
    raw_angles = np.degrees(np.arctan2(dy, dx))  # -180..+180
    # 0 = vertical up, 180 = vertical down
    orient = (raw_angles + 90) % 180
    mask = (orient <= angle_tolerance) | (orient >= 180 - angle_tolerance)
    vertical = lines[mask]
    if vertical.shape[0] < 2:
        print("Could not find two vertical lines within tolerance.")
        return original_img

    # Create a mask for lines where both y-coordinates are in the top half
    # A line (x1,y1,x2,y2) is in the top half if both y1 and y2 are less than y_midpoint

    y_midpoint = abs(h * bottom_verticals_threshold)
    top_half_mask = (vertical[:, 0, 1] < y_midpoint) & (
        vertical[:, 0, 3] < y_midpoint)
    vertical = vertical[top_half_mask]

    if vertical.shape[0] < 2:
        print(
            "Could not find two vertical lines in the top half of the image within tolerance."
        )
        return original_img
    # Extract endpoints
    vertical = vertical[:, 0]  # shape (N,4)

    # min_dist = int(w * min_dist_pct)

    # 4. Select two longest vertical lines

    def length_sq(l):
        return (l[2] - l[0]) ** 2 + (l[3] - l[1]) ** 2

    sorted_lines = sorted(vertical, key=length_sq, reverse=True)
    min_dist = 10  # your minimum x-distance in pixels

    # find the longest pair whose centers are >= min_dist apart
    for i, l1 in enumerate(sorted_lines):
        x1 = (l1[0] + l1[2]) // 2
        for l2 in sorted_lines[i + 1:]:
            x2 = (l2[0] + l2[2]) // 2
            if abs(x2 - x1) >= min_dist:
                # we found a valid pair
                break
        else:
            # no valid partner for l1, continue with next
            continue
        # break out of outer loop as well
        break
    else:
        print(f"Could not find two vertical lines ≥ {min_dist}px apart.")
        return original_img
    # 5. Compute bounding coords
    x1 = (l1[0] + l1[2]) // 2
    x2 = (l2[0] + l2[2]) // 2
    left_x, right_x = int(min(x1, x2)), int(max(x1, x2))
    ys = [l1[1], l1[3], l2[1], l2[3]]
    top_y, bottom_y = int(min(ys)), int(max(ys))

    # 6. Pad and clamp
    h, w = gray.shape[:2]
    left = max(0, left_x - padding)
    right = min(w, right_x + padding)
    top = max(0, top_y - padding)
    bottom = min(h, bottom_y + padding)

    # 8. Crop and return
    img = img[top:bottom, left:right]

    return img


def image_crop(image: np.ndarray, points):
    # return image[min(y):max(y), min(x),max(x)]
    return image[
        int(points[0][1]): int(points[1][1]), int(points[0][0]): int(points[1][0])
    ]


def set_surface_line(
    experimental_drop: ExperimentalDrop, experimental_setup: ExperimentalSetup
) -> None:
    # message = []

    #
    # if experimental_drop.cropped_image is None:
    #     if experimental_setup.drop_id_method == "User-selected":
    #         msgbox.showwarning("Warning", "Please select the drop region")
    #         set_drop_region(experimental_drop, experimental_setup)
    #         return
    #     # autuomatic
    #     else:
    #         set_drop_region(experimental_drop, experimental_setup)

    # if experimental_setup.threshold_method == "User-selected":
    #     if experimental_setup.threshold_val is None:
    #         threshold = simpledialog.askinteger("Input Required", "Enter the threshold value:")
    #         if threshold is None:  # User pressed "Cancel"
    #             msgbox.showwarning("Warning", "Threshold is required to continue.")
    #             return
    #         experimental_setup.threshold_val = threshold

    # extract_drop_profile(experimental_drop, experimental_setup)

    if experimental_setup.baseline_method == ThresholdSelect.AUTOMATED:
        experimental_drop.drop_contour, experimental_drop.contact_points = (
            prepare_hydrophobic(experimental_drop.contour)
        )
    elif experimental_setup.baseline_method == ThresholdSelect.USER_SELECTED:
        user_line(experimental_drop, experimental_setup)


def correct_tilt(
    experimental_drop: ExperimentalDrop, experimental_setup: ExperimentalSetup
) -> None:
    if experimental_setup.baseline_method == ThresholdSelect.AUTOMATED:
        experimental_drop.cropped_image = tilt_correction(
            experimental_drop.cropped_image, experimental_drop.contact_points
        )

    # gets tricky where the baseline is manually set because under the current workflow users would
    # be required to re-input their baseline until it's flat - when the baseline should be flat
    # and known once it's set and corrected for
    elif experimental_setup.baseline_method == ThresholdSelect.USER_SELECTED:
        rotated_img_crop = tilt_correction(
            img, experimental_drop.contact_points, user_set_baseline=True
        )


def set_scale(image_size: Tuple[int, int, int], screen_size: List[int]) -> float:
    x_ratio = image_size[1] / float(screen_size[0])
    y_ratio = image_size[0] / float(screen_size[1])
    max_ratio = max(x_ratio, y_ratio)
    scale = 1
    if max_ratio > MAX_IMAGE_TO_SCREEN_RATIO:
        scale = MAX_IMAGE_TO_SCREEN_RATIO / max_ratio
    return scale


def set_screen_position(screen_size: List[int]) -> List[int]:
    # percentage room free
    percent_free_space = 0.5 * (1 - MAX_IMAGE_TO_SCREEN_RATIO)
    x_position = int(percent_free_space * screen_size[0])
    # 0.5 moves window a little bit higher
    y_position = int(0.5 * percent_free_space * screen_size[1])
    return [x_position, y_position]


# , line_colour=(0, 0, 255), line_thickness=2):
def user_roi(
    raw_image: np.ndarray, title: str, scale: float, screen_position
) -> List[Tuple[float, float]]:
    """
    Function to select a region of interest (ROI) in an image using mouse events.
    User can click and drag to create a rectangle, and press Enter to confirm the selection.
    """
    global drawing
    global ix, iy
    global fx, fy
    global image_TEMP
    global img
    # raw_image = raw_image2
    # raw_image = np.flipud(cv2.cvtColor(raw_image2,cv2.COLOR_GRAY2BGR))
    # raw_image = np.flipud(raw_image2)
    drawing = False  # true if mouse is pressed
    ix, iy = -1, -1
    fx, fy = -1, -1

    cv2.namedWindow(title, cv2.WINDOW_AUTOSIZE)
    cv2.moveWindow(title, screen_position[0], screen_position[1])
    cv2.setMouseCallback(title, draw_rectangle)
    # scale =1
    image_TEMP = cv2.resize(raw_image, (0, 0), fx=scale, fy=scale)

    img = image_TEMP.copy()

    while 1:
        # Add message to the image before showing
        img_display = img.copy()
        msg = "Drag to select a region. Press Enter to confirm."
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 0.6
        color = (255, 255, 255)
        thickness = 2
        # Position: bottom left, 10px from left, 10px from bottom
        text_size, _ = cv2.getTextSize(msg, font, font_scale, thickness)
        text_x = 10
        text_y = img_display.shape[0] - 10
        cv2.putText(
            img_display,
            msg,
            (text_x, text_y),
            font,
            font_scale,
            color,
            thickness,
            cv2.LINE_AA,
        )

        cv2.imshow(title, img_display)

        k = cv2.waitKey(1) & 0xFF
        if k != 255:
            if (k == 13) or (k == 32):
                # either 'return' or 'space' pressed
                # break
                if ((fx - ix) * (fy - iy)) != 0:  # ensure there is an enclosed region
                    break
            if k == 27:
                # 'esc'
                kill()

    cv2.destroyAllWindows()
    min_x = min(ix, fx) / scale
    max_x = max(ix, fx) / scale
    min_y = min(iy, fy) / scale
    max_y = max(iy, fy) / scale
    return [(min_x, min_y), (max_x, max_y)]


def user_line(
    experimental_drop: ExperimentalDrop, experimental_setup: ExperimentalSetup
) -> None:
    # scale = set_scale(experimental_drop.image.shape, experimental_setup.screen_resolution)
    screen_position = set_screen_position(experimental_setup.screen_resolution)
    raw_image = experimental_drop.cropped_image
    drop_data = experimental_drop.contour.astype(float)
    CPs = experimental_drop.contact_points
    title = "Define surface line"

    # line = experimental_drop.surface_data # not set yet
    region = experimental_setup.drop_region

    global drawing
    global ix, iy
    global fx, fy
    global image_TEMP
    global img

    DRAW_TANGENT_LINE_WHILE_SETTING_BASELINE = True
    TEMP = False
    baseline_def_method = "use user-inputted points"

    # raw_image = raw_image2
    # raw_image = np.flipud(cv2.cvtColor(raw_image2,cv2.COLOR_GRAY2BGR))
    # raw_image = np.flipud(raw_image2)
    drawing = True  # true if mouse is pressed
    ix, iy = -1, -1
    fx, fy = -1, -1

    region = np.floor(region)
    # print(region)

    # print(region[0,0])
    # print(region[1,0])
    # print(region[0,1])
    # print(region[1,1])

    #    cv2.setMouseCallback(title, draw_line)

    scale = 1
    if TEMP:
        image_TEMP = cv2.resize(
            raw_image[
                int(region[0, 1]): int(region[1, 1]),
                int(region[0, 0]): int(region[1, 0]),
            ],
            (0, 0),
            fx=scale,
            fy=scale,
        )
    else:
        image_TEMP = raw_image.copy()
    img = image_TEMP.copy()

    # set surface line starting estimate
    N = np.shape(drop_data)[0]
    A = 1  # 50 # maybe lower this?
    xx = np.concatenate((drop_data[0:A, 0], drop_data[N - A: N + 1, 0]))
    yy = np.concatenate((drop_data[0:A, 1], drop_data[N - A: N + 1, 1]))
    coefficients = np.polyfit(xx, yy, 1)
    line = np.poly1d(coefficients)

    xx = np.array([0, img.shape[1]])
    yy = line(xx)  # gives a starting guess for the line position

    ix0, fx0 = xx.astype(int)
    iy0, fy0 = yy.astype(int)

    ix, fx = ix0, fx0
    iy, fy = iy0, fy0

    cv2.namedWindow(title, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(title, 500, 400)
    cv2.moveWindow(title, screen_position[0], screen_position[1])
    try:
        cv2.setWindowProperty(title, 5, 0)  # Lock size (if supported)
    except:
        pass  # Safe fallback for macOS or older OpenCV
    cv2.moveWindow(title, screen_position[0], screen_position[1])

    # so that things can be drawn over the image which surface line is changed
    if DRAW_TANGENT_LINE_WHILE_SETTING_BASELINE:
        conans = {}
        if 0:
            for i, n in enumerate(drop_data):
                if n[0] == CPs[0][0] and int(n[1]) == int(CPs[0][1]):
                    start_index = i
                if int(n[0]) == int(CPs[1][0]) and int(n[1]) == int(CPs[1][1]):
                    end_index = i
            auto_drop = drop_data.copy()[start_index:end_index]
        else:
            auto_drop = drop_data

    # Add guidance for the user
    print("\nKeyboard Controls:")
    print("W = Move Up")
    print("S = Move Down")
    print("A = Rotate Left")
    print("D = Rotate Right")
    print("O = Reset to Original Position")
    print("P = Print Debug Info")
    print("ENTER/SPACE = Confirm Selection")
    print("ESC = Exit\n")

    while 1:
        img = cv2.resize(img, (500, 400))
        cv2.imshow(title, img)
        # cv2.circle(img,(200,200),5,(255,255,0),2)
        # line_colour,line_thickness)
        cv2.line(img, (ix, iy), (fx, fy), (0, 255, 0), 2)
        # Plot pixels above line
        # cv2.waitKey(0)
        v1 = (ix - fx, iy - fy)  # (1,coefficients[1])   # Vector 1

        # print(np.shape(drop_data))
        # print(drop_data)

        # Plot pixels above line
        v1 = (ix - fx, iy - fy)

        if 1:
            drop = []
            for i in drop_data:
                cx, cy = i
                v2 = (cx - ix, cy - iy)  # Vector 1
                xp = v1[0] * v2[1] - v1[1] * v2[0]  # Cross product
                if xp > 0:
                    drop.append([cx, cy])
                    cv2.circle(img, (int(cx), int(cy)), 2, (255, 255, 255), 1)
        else:
            drop = []
            for i in drop_data:
                cx, cy = i
                # if contour point y value less than line y value
                if cy < line(cx):
                    drop.append([cx, cy])
                    cv2.circle(img, (int(cx), int(cy)), 2, (255, 255, 255), 1)

        # drop is the contour above the user-inputted line
        drop = np.asarray(drop).astype(float)

        if 0:
            plt.imshow(img)
            plt.title("check contour after being cut by baseline")
            plt.plot(drop[:, 0], drop[:, 1])
            plt.show()
            plt.close()

        experimental_drop.drop_contour = drop
        CPs = {0: drop[0], 1: drop[-1]}
        experimental_drop.contact_points = CPs

        if DRAW_TANGENT_LINE_WHILE_SETTING_BASELINE:
            methods_boole = experimental_setup.analysis_methods_ca
            if (
                methods_boole[FittingMethod.TANGENT_FIT]
                or methods_boole[FittingMethod.POLYNOMIAL_FIT]
                or methods_boole[FittingMethod.CIRCLE_FIT]
                or methods_boole[FittingMethod.ELLIPSE_FIT]
            ):
                from modules.fitting.fits import perform_fits

                perform_fits(
                    experimental_drop,
                    tangent=methods_boole[FittingMethod.TANGENT_FIT],
                    polynomial=methods_boole[FittingMethod.POLYNOMIAL_FIT],
                    circle=methods_boole[FittingMethod.CIRCLE_FIT],
                    ellipse=methods_boole[FittingMethod.ELLIPSE_FIT],
                )
            if methods_boole[FittingMethod.TANGENT_FIT]:
                tangent_lines = tuple(
                    experimental_drop.contact_angles["tangent fit"]["tangent lines"]
                )
                cv2.line(
                    img,
                    (int(tangent_lines[0][0][0]), int(tangent_lines[0][0][1])),
                    (int(tangent_lines[0][1][0]), int(tangent_lines[0][1][1])),
                    (0, 0, 255),
                    2,
                )
                cv2.line(
                    img,
                    (int(tangent_lines[1][0][0]), int(tangent_lines[1][0][1])),
                    (int(tangent_lines[1][1][0]), int(tangent_lines[1][1][1])),
                    (0, 0, 255),
                    2,
                )
            if (
                methods_boole[FittingMethod.POLYNOMIAL_FIT] == True
                and not methods_boole[FittingMethod.TANGENT_FIT]
            ):
                tangent_lines = tuple(
                    experimental_drop.contact_angles["polynomial fit"]["tangent lines"]
                )
                cv2.line(img, tangent_lines[0][0],
                         tangent_lines[0][1], (0, 0, 255), 2)
                cv2.line(img, tangent_lines[1][0],
                         tangent_lines[1][1], (0, 0, 255), 2)
            if methods_boole[FittingMethod.CIRCLE_FIT]:
                xc, yc = experimental_drop.contact_angles["circle fit"]["circle center"]
                r = experimental_drop.contact_angles["circle fit"]["circle radius"]
                cv2.circle(img, (int(xc), int(yc)), int(r), (255, 150, 0), 1)
            if methods_boole[FittingMethod.ELLIPSE_FIT]:
                center = experimental_drop.contact_angles["ellipse fit"][
                    "ellipse center"
                ]
                axes = experimental_drop.contact_angles["ellipse fit"][
                    "ellipse a and b"
                ]
                phi = experimental_drop.contact_angles["ellipse fit"][
                    "ellipse rotation"
                ]
                cv2.ellipse(
                    img,
                    (int(center[0]), int(center[1])),
                    (int(axes[0]), int(axes[1])),
                    phi,
                    0,
                    360,
                    (0, 88, 255),
                    1,
                )

        # Get key press with a simpler approach
        k = cv2.waitKey(1) & 0xFF

        # Process key presses using WASD controls instead of arrow keys
        if k == 13 or k == 32:  # ENTER / SPACE
            if ((fx - ix) * (fy - iy)) != 0:  # Ensure enclosed region
                break
            else:
                print("something is not right...")
                print(fx, ix, fy, iy, (fx - ix) * (fy - iy))
                break

        elif k == 27:  # ESC
            kill()

        elif k == ord("s"):  # 's' for DOWN
            print("DOWN key pressed (s)")
            fy += 1
            iy += 1

        elif k == ord("w"):  # 'w' for UP
            print("UP key pressed (w)")
            fy -= 1
            iy -= 1

        elif k == ord("o") or k == ord("O"):  # 'o' key
            print("O key pressed")
            fx, fy = fx0, fy0
            ix, iy = ix0, iy0

        elif k == ord("a"):  # 'a' for LEFT rotation
            print("LEFT key pressed (a)")
            x0 = np.array([ix, iy])
            x1 = np.array([fx, fy])
            xc = 0.5 * (x0 + x1)
            theta = 0.1 / 180 * np.pi
            theta = -theta  # counter-clockwise

            rotation = np.array(
                [[np.cos(theta), -np.sin(theta)],
                 [np.sin(theta), np.cos(theta)]]
            )
            x0r = rotation @ (x0 - xc).T + xc
            x1r = rotation @ (x1 - xc).T + xc

            ix, iy = x0r.astype(int)
            fx, fy = x1r.astype(int)

        elif k == ord("d"):  # 'd' for RIGHT rotation
            print("RIGHT key pressed (d)")
            x0 = np.array([ix, iy])
            x1 = np.array([fx, fy])
            xc = 0.5 * (x0 + x1)
            theta = 0.1 / 180 * np.pi
            # No negation for clockwise rotation

            rotation = np.array(
                [[np.cos(theta), -np.sin(theta)],
                 [np.sin(theta), np.cos(theta)]]
            )
            x0r = rotation @ (x0 - xc).T + xc
            x1r = rotation @ (x1 - xc).T + xc

            ix, iy = x0r.astype(int)
            fx, fy = x1r.astype(int)

        elif k == ord("p") or k == ord("P"):  # 'p' key
            for key in conans.keys():
                print(key, ":", conans[key])
            print()

        # Redraw the image after any update
        if TEMP:
            image_TEMP = cv2.resize(
                raw_image[
                    int(region[0, 1]): int(region[1, 1]),
                    int(region[0, 0]): int(region[1, 0]),
                ],
                (0, 0),
                fx=scale,
                fy=scale,
            )
        else:
            image_TEMP = raw_image.copy()

        img = image_TEMP.copy()
        # cv2.line(img, (ix, iy), (fx, fy), (0, 255, 0), 2)
        cv2.putText(
            img,
            "Use W/A/S/D to move. ESC = cancel.",
            (10, img.shape[0] - 15),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.3,
            (255, 255, 255),
            1,
            cv2.LINE_AA,
        )

        cv2.putText(
            img,
            "Enter/Space = confirm.",
            (10, img.shape[0] - 5),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.3,
            (255, 255, 255),
            1,
            cv2.LINE_AA,
        )
        cv2.line(img, (ix, iy), (fx, fy), (0, 255, 0), 2)

    cv2.destroyAllWindows()
    min_x = min(ix, fx) / scale
    max_x = max(ix, fx) / scale
    min_y = min(iy, fy) / scale
    max_y = max(iy, fy) / scale


def run_set_surface_line(
    experimental_drop: ExperimentalDrop,
    experimental_setup: ExperimentalSetup,
    result_queue,
):

    set_surface_line(experimental_drop, experimental_setup)
    result_queue.put(experimental_drop.contact_angles)


# mouse callback function


def draw_rectangle(event: int, x, y, flags, param):
    global ix, iy, drawing
    global fx, fy
    global image_TEMP
    global img

    if event == cv2.EVENT_LBUTTONDOWN:
        img = image_TEMP.copy()
        drawing = True
        ix, iy = x, y

    elif event == cv2.EVENT_MOUSEMOVE:
        if drawing == True:
            img = image_TEMP.copy()
            # line_colour,line_thickness)
            cv2.rectangle(img, (ix, iy), (x, y), (0, 0, 255), 2)

    elif event == cv2.EVENT_LBUTTONUP:
        img = image_TEMP.copy()
        drawing = False
        fx, fy = x, y
        # line_colour,line_thickness)
        cv2.rectangle(img, (ix, iy), (fx, fy), (0, 255, 0), 2)


# mouse callback function


def draw_line(event: int, x, y, flags, param):
    global ix, iy, drawing
    global fx, fy
    global image_TEMP
    global img

    if event == cv2.EVENT_LBUTTONDOWN:
        img = image_TEMP.copy()
        drawing = True
        ix, iy = x, y

    elif event == cv2.EVENT_MOUSEMOVE:
        if drawing == True:
            img = image_TEMP.copy()
            # line_colour,line_thickness)
            cv2.line(img, (ix, iy), (x, y), (0, 0, 255), 2)

    elif event == cv2.EVENT_LBUTTONUP:
        img = image_TEMP.copy()
        drawing = False
        fx, fy = x, y
        # line_colour,line_thickness)
        cv2.line(img, (ix, iy), (fx, fy), (0, 255, 0), 2)


def kill():
    sys.exit()


def distance(p1, p2):
    """This function computes the distance between 2 points defined by
    P1 = (x1,y1) and P2 = (x2,y2)"""
    return ((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2) ** 0.5


def optimized_path(coords, start=None):
    """This function finds the nearest point to a point
    coords should be a list in this format coords = [ [x1, y1], [x2, y2] , ...]
    https://stackoverflow.com/questions/45829155/sort-points-in-order-to-have-a-continuous-curve-using-python
    """
    if isinstance(coords, list) == False:
        coords = coords.tolist()
    if 0:
        if isinstance(start, list) == False:
            try:
                start = start.tolist()
            except:
                start = list(start)
    if start is None:
        start = coords[0]
    pass_by = coords
    path = [start]
    pass_by.remove(start)
    while pass_by:
        nearest = min(pass_by, key=lambda x: distance(path[-1], x))
        path.append(nearest)
        pass_by.remove(nearest)
    path = np.array(path)
    return path


def intersection(center, radius, p1, p2):
    """find the two points where a secant intersects a circle"""

    dx, dy = p2[0] - p1[0], p2[1] - p1[1]

    a = dx**2 + dy**2
    b = 2 * (dx * (p1[0] - center[0]) + dy * (p1[1] - center[1]))
    c = (p1[0] - center[0]) ** 2 + (p1[1] - center[1]) ** 2 - radius**2

    discriminant = b**2 - 4 * a * c
    assert discriminant > 0, "Not a secant!"

    t1 = (-b + discriminant**0.5) / (2 * a)
    t2 = (-b - discriminant**0.5) / (2 * a)

    return (dx * t1 + p1[0], dy * t1 + p1[1]), (dx * t2 + p1[0], dy * t2 + p1[1])


def ML_prepare_hydrophobic(coords_in):
    coords = coords_in
    coords[:, 1] = -coords[:, 1]  # flip
    # print('length of coords: ',len(coords))

    # isolate the top of the contour so excess surface can be deleted
    percent = 0.1
    bottom = []
    top = []  # will need this later
    div_line_value = (
        min(coords[:, [1]]) + (max(coords[:, [1]]) -
                               min(coords[:, [1]])) * percent
    )
    for n in coords:
        if n[1] < div_line_value:
            bottom.append(n)
        else:
            top.append(n)

    bottom = np.array(bottom)
    top = np.array(top)

    del_indexes = []
    for index, coord in enumerate(coords):
        if coord[0] > max(top[:, 0]) or coord[0] < min(top[:, 0]):
            del_indexes.append(index)
    # halfdrop = np.delete(halfdrop,del_indexes)
    coords = np.delete(coords, del_indexes, axis=0)

    if 0:
        plt.title("isolated coords, length: " + str(len(coords)))
        plt.plot(coords[:, 0], coords[:, 1])
        plt.show()
        plt.close()

    # find the apex of the drop and split the contour into left and right sides

    xtop, ytop = top[:, 0], top[:, 1]  # isolate top 90% of drop

    xapex = (max(xtop) + min(xtop)) / 2
    # yapex = max(ytop)
    # coords[:,1] = -coords[:,1]

    l_drop = []
    r_drop = []
    for n in coords:
        if n[0] < xapex:
            l_drop.append(n)
        if n[0] > xapex:
            r_drop.append(n)
    l_drop = np.array(l_drop)
    r_drop = np.array(r_drop)

    # print('length of left drop is: ',len(l_drop))
    # print('length of right drop is: ', len(r_drop))

    # transpose both half drops so that they both face right and the apex of both is at 0,0
    # r_drop[:,[0]] = r_drop[:,[0]] - min(r_drop[:,[0]])
    # l_drop[:,[0]] = -l_drop[:,[0]] + max(l_drop[:,[0]])
    r_drop[:, [0]] = r_drop[:, [0]] - xapex
    l_drop[:, [0]] = -l_drop[:, [0]] + xapex

    counter = 0
    CV_contours = {}

    for halfdrop in [l_drop, r_drop]:
        if halfdrop[0, 1] < halfdrop[-1, 1]:
            halfdrop = halfdrop[::-1]

        X = halfdrop[:, 0]
        Z = halfdrop[:, 1]

        lowest = min(Z)
        Z = Z + abs(lowest)

        X = X / max(Z)
        Z = Z / max(Z)

        # zero padd contours to
        coordinates = []
        input_len = 1100
        len_cont = len(X)

        # if len(X) > global_max_len:
        #    global_max_len = len(X)

        if len(X) > input_len:
            print(len(X))
            raise Exception(
                "Contour of length "
                + str(len(X))
                + " is too long for the designated output dimensionality of ("
                + str(input_len)
                + ",2)"
            )

        for i in range(input_len):
            if i < len(X):
                a = X[i]
                b = Z[i]
                coord = [a, b]
                coordinates.append(coord)
            else:
                coordinates.append([0, 0])
        if 0:
            jet = plt.get_cmap("jet")
            colors = iter(jet(np.linspace(0, 1, len(coordinates))))
            for k in coordinates:
                plt.plot(k[0], k[1], "o", color=next(colors))
            plt.title("Halfdrop")
            plt.show()
            plt.close()
        # key = image.split('/')[-1].split('_')[-1][:-4]
        key = counter
        CV_contours[key] = np.array(coordinates)

        counter += 1

    pred_ds = np.zeros((2, input_len, 2))
    for counter in [0, 1]:
        pred_ds[counter] = CV_contours[counter]

    return pred_ds



from modules.image.read_image import (
    get_image,
    save_image,
    import_from_source,
    image_from_Flea3,
    image_from_harddrive,
    get_import_filename,
    image_from_camera,
    grabanimage,
)

from unittest.mock import patch, MagicMock
import pytest
import numpy as np
import sys
import os


sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


# Fixture to create a mock ExperimentalDrop object
@pytest.fixture
def mock_experimental_drop():
    class ExperimentalDrop:
        image = None
        time = None

    return ExperimentalDrop()


# Fixture to create a mock ExperimentalSetup object


@pytest.fixture
def mock_experimental_setup():
    class ExperimentalSetup:
        image_source = "Local images"
        directory_string = "test_directory"
        filename = "test_image.png"
        import_files = ["test_image.png"]
        save_images_boole = True
        create_folder_boole = True
        time_string = "2024-09-08-120000"

    return ExperimentalSetup()


# Test get_image function


@patch("cv2.imwrite")
@patch("os.makedirs")
def test_get_image(
    mock_makedirs, mock_imwrite, mock_experimental_drop, mock_experimental_setup
):
    get_image(mock_experimental_drop, mock_experimental_setup, 0)
    mock_makedirs.assert_called_once()

    # Reset the mock so we only check the next call
    mock_makedirs.reset_mock()

    # Test without create_folder_boole
    mock_experimental_setup.create_folder_boole = False
    get_image(mock_experimental_drop, mock_experimental_setup, 0)
    mock_makedirs.assert_not_called()


# Test save_image function


def test_save_image(mock_experimental_drop, mock_experimental_setup):
    mock_experimental_drop.image = np.zeros((100, 100, 3), dtype=np.uint8)
    with patch("cv2.imwrite") as mock_imwrite:
        save_image(mock_experimental_drop, mock_experimental_setup, 1)
        mock_imwrite.assert_called_once()


# Test import_from_source with local images


@patch("modules.image.read_image.image_from_harddrive")
def test_import_from_source_local(
    mock_image_from_harddrive, mock_experimental_drop, mock_experimental_setup
):
    import_from_source(mock_experimental_drop, mock_experimental_setup, 0)
    mock_image_from_harddrive.assert_called_once()


# Test image_from_Flea3 function


@patch("subprocess.call")
@patch("cv2.imread")
def test_image_from_Flea3(mock_imread, mock_subprocess, mock_experimental_drop):
    mock_imread.return_value = np.zeros((100, 100, 3), dtype=np.uint8)
    image_from_Flea3(mock_experimental_drop)
    mock_subprocess.assert_called_once_with(["./FCGrab"])
    mock_imread.assert_called_once_with("FCG.pgm", 1)


# Test image_from_harddrive function


def test_image_from_harddrive(mock_experimental_drop, mock_experimental_setup):
    with patch(
        "cv2.imread", return_value=np.zeros((100, 100, 3), dtype=np.uint8)
    ) as mock_imread:
        image_from_harddrive(mock_experimental_drop, mock_experimental_setup, 0)
        mock_imread.assert_called_once_with("test_image.png", 1)


# Test get_import_filename function


def test_get_import_filename(mock_experimental_setup):
    filename = get_import_filename(mock_experimental_setup, 0)
    assert filename == "test_image.png"


# Test image_from_camera function


@patch("modules.image.read_image.grabanimage")
def test_image_from_camera(mock_grabanimage, mock_experimental_drop):
    with patch(
        "cv2.imread", return_value=np.zeros((100, 100, 3), dtype=np.uint8)
    ) as mock_imread:
        image_from_camera(mock_experimental_drop)
        mock_grabanimage.assert_called_once()
        mock_imread.assert_called_once_with("USBtemp.png", 1)


# Test grabanimage function


@patch("cv2.VideoCapture")
def test_grabanimage(mock_VideoCapture):
    mock_camera = MagicMock()
    mock_VideoCapture.return_value = mock_camera
    mock_camera.read.return_value = (True, np.zeros((100, 100, 3), dtype=np.uint8))
    with patch("cv2.imwrite") as mock_imwrite:
        grabanimage()
        mock_camera.read.assert_called()
        mock_imwrite.assert_called_once_with(
            "USBtemp.png", mock_camera.read.return_value[1]
        )


# Run the tests when this script is executed directly
if __name__ == "__main__":
    pytest.main()



from modules.image.select_regions import (
    set_drop_region,
    set_surface_line,
    correct_tilt,
    user_line,
    draw_rectangle,
    draw_line,
    optimized_path,
    intersection,
    ML_prepare_hydrophobic,
)
from modules.core.classes import ExperimentalSetup, ExperimentalDrop
from modules.preprocessing.preprocessing import prepare_hydrophobic
from modules.fitting.fits import perform_fits
from utils.enums import RegionSelect, ThresholdSelect

from itertools import cycle
from unittest.mock import patch, MagicMock
from numpy.testing import assert_array_equal
import numpy as np
import unittest
import cv2
import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


class TestSelectRegions(unittest.TestCase):

    @patch("modules.preprocessing.preprocessing.auto_crop")
    def test_set_drop_region_auto(self, mock_auto_crop: MagicMock):
        # Create mock return value
        # Simulate a blank image of size 100x100
        mock_image = np.zeros((100, 100, 3))
        mock_auto_crop.return_value = (
            mock_image,
            (10, 90, 20, 80),
        )  # Simulate cropping result

        # Simulate experimental_drop and experimental_setup objects
        experimental_drop = ExperimentalDrop()
        experimental_drop.image = mock_image
        experimental_setup = ExperimentalSetup()
        experimental_setup.drop_id_method = RegionSelect.AUTOMATED
        experimental_setup.screen_resolution = [1, 1]

        # Call the function to be tested
        set_drop_region(experimental_drop, experimental_setup, index=0)

        # Check if auto_crop was called correctly
        mock_auto_crop.assert_called_once_with(mock_image)

        # Verify if experimental_drop handled auto_crop's return value correctly
        self.assertTrue(np.array_equal(experimental_drop.cropped_image, mock_image))
        self.assertEqual(experimental_setup.drop_region, [(10, 20), (90, 80)])

    @patch("modules.image.select_regions.user_roi")
    @patch("modules.image.select_regions.image_crop")
    def test_set_drop_region_user_selected(
        self, mock_image_crop: MagicMock, mock_user_roi: MagicMock
    ):
        # Create mock return value
        # Simulate a blank image of size 100x100
        mock_image = np.zeros((100, 100, 3))
        # Simulate the user-selected region
        mock_user_roi.return_value = [(10, 20), (90, 80)]

        # Simulate the cropped image as a 3D array
        # Assume the cropped result is still 3D
        mock_cropped_image = np.zeros((60, 80, 3))
        mock_image_crop.return_value = mock_cropped_image

        # Simulate experimental_drop and experimental_setup objects
        experimental_drop = ExperimentalDrop()
        experimental_drop.image = mock_image
        experimental_setup = ExperimentalSetup()
        experimental_setup.drop_id_method = RegionSelect.USER_SELECTED
        experimental_setup.screen_resolution = [1, 1]

        # Call the function to be tested
        set_drop_region(experimental_drop, experimental_setup, index=1)

        # Check if user_roi was called correctly
        mock_user_roi.assert_called_once_with(
            mock_image, "Select drop region for Image 1", 0.008, [0, 0]
        )

        # Verify if image_crop was called correctly
        mock_image_crop.assert_called_once_with(mock_image, [(10, 20), (90, 80)])

        # Check if experimental_drop and experimental_setup handled user_roi's return value correctly
        self.assertEqual(experimental_setup.drop_region, [(10, 20), (90, 80)])
        self.assertEqual(
            experimental_drop.cropped_image.shape, (60, 80, 3)
        )  # Verify the cropped image size


class TestSurfaceLine(unittest.TestCase):

    @patch("modules.image.select_regions.prepare_hydrophobic")
    @patch("modules.preprocessing.preprocessing.optimized_path")
    def test_set_surface_line_auto(self, mock_optimized_path, mock_prepare_hydrophobic):
        # Simulate optimized_path return value to ensure enough data points for distance calculation
        mock_optimized_path.return_value = np.array(
            [[0, 0], [1, 2], [2, 4], [3, 6], [4, 8], [5, 10]]
        )

        # Create mock return values
        mock_contour = np.array(
            [[0, 0], [1, 2], [2, 4], [3, 6], [4, 8], [5, 10]]
        )  # Using more contour points
        mock_contact_points = {0: [0, 0], 1: [5, 10]}
        mock_prepare_hydrophobic.return_value = (mock_contour, mock_contact_points)

        # Simulate experimental_drop and experimental_setup objects
        experimental_drop = MagicMock()
        experimental_drop.contour = mock_contour
        experimental_setup = MagicMock()
        experimental_setup.baseline_method = ThresholdSelect.AUTOMATED

        # Call the function to be tested
        set_surface_line(experimental_drop, experimental_setup)

        # Check if prepare_hydrophobic was called correctly
        mock_prepare_hydrophobic.assert_called_once_with(mock_contour)

        # Verify if the return values were handled correctly
        self.assertTrue(np.array_equal(experimental_drop.drop_contour, mock_contour))
        self.assertEqual(experimental_drop.contact_points, mock_contact_points)


class TestCorrectTilt(unittest.TestCase):

    @patch("modules.image.select_regions.tilt_correction")
    def test_correct_tilt_auto(self, mock_tilt_correction):
        # Create mock return value
        mock_cropped_image = np.zeros((100, 100, 3))
        mock_tilt_correction.return_value = mock_cropped_image

        # Simulate experimental_drop and experimental_setup objects
        experimental_drop = MagicMock()
        experimental_drop.cropped_image = mock_cropped_image
        experimental_drop.contact_points = [(10, 20), (90, 80)]
        experimental_setup = MagicMock()
        experimental_setup.baseline_method = ThresholdSelect.AUTOMATED

        # Call the function to be tested
        correct_tilt(experimental_drop, experimental_setup)

        # Check if tilt_correction was called correctly
        mock_tilt_correction.assert_called_once_with(
            mock_cropped_image, [(10, 20), (90, 80)]
        )

        # Verify if the return values were handled correctly
        self.assertTrue(
            np.array_equal(experimental_drop.cropped_image, mock_cropped_image)
        )


class TestDrawRectangle(unittest.TestCase):

    @patch("modules.image.select_regions.cv2.rectangle")
    @patch.dict(
        "modules.image.select_regions.__dict__",
        {"image_TEMP": np.zeros((100, 100, 3)), "img": np.zeros((100, 100, 3))},
    )
    def test_draw_rectangle(self, mock_rectangle):
        global drawing, ix, iy, fx, fy
        # Initialize local variables
        drawing = False
        ix, iy = 0, 0
        fx, fy = 0, 0

        # Simulate mouse button down event
        draw_rectangle(cv2.EVENT_LBUTTONDOWN, 10, 20, None, None)
        # No rectangle should be drawn when mouse is pressed down
        mock_rectangle.assert_not_called()

        # Simulate mouse move event, cv2.rectangle should be called
        draw_rectangle(cv2.EVENT_MOUSEMOVE, 30, 40, None, None)
        called_args = mock_rectangle.call_args
        assert np.array_equal(
            called_args[0][0], np.zeros((100, 100, 3))
        )  # Manually compare numpy arrays
        assert called_args[0][1] == (10, 20)
        assert called_args[0][2] == (30, 40)
        assert called_args[0][3] == (0, 0, 255)
        assert called_args[0][4] == 2

        # Simulate mouse button release event, cv2.rectangle should be called again
        draw_rectangle(cv2.EVENT_LBUTTONUP, 50, 60, None, None)
        called_args = mock_rectangle.call_args
        assert np.array_equal(
            called_args[0][0], np.zeros((100, 100, 3))
        )  # Manually compare numpy arrays
        assert called_args[0][1] == (10, 20)
        assert called_args[0][2] == (50, 60)
        assert called_args[0][3] == (0, 255, 0)
        assert called_args[0][4] == 2


class TestDrawLine(unittest.TestCase):

    @patch("modules.image.select_regions.cv2.line")
    @patch.dict(
        "modules.image.select_regions.__dict__",
        {"image_TEMP": np.zeros((100, 100, 3)), "img": np.zeros((100, 100, 3))},
    )
    def test_draw_line(self, mock_line):
        global drawing, ix, iy, fx, fy
        # Initialize local variables
        drawing = False
        ix, iy = 0, 0
        fx, fy = 0, 0

        # Simulate mouse button down event
        draw_line(cv2.EVENT_LBUTTONDOWN, 10, 20, None, None)
        mock_line.assert_not_called()  # No line should be drawn when mouse is pressed down

        # Simulate mouse move event, cv2.line should be called
        draw_line(cv2.EVENT_MOUSEMOVE, 30, 40, None, None)
        args, kwargs = mock_line.call_args
        # Compare first argument
        assert_array_equal(args[0], np.zeros((100, 100, 3)))
        self.assertEqual(args[1], (10, 20))  # Compare start point
        self.assertEqual(args[2], (30, 40))  # Compare end point
        self.assertEqual(args[3], (0, 0, 255))  # Compare color
        self.assertEqual(args[4], 2)  # Compare line thickness

        # Simulate mouse button release event, cv2.line should be called again
        draw_line(cv2.EVENT_LBUTTONUP, 50, 60, None, None)
        args, kwargs = mock_line.call_args
        # Compare first argument
        assert_array_equal(args[0], np.zeros((100, 100, 3)))
        self.assertEqual(args[1], (10, 20))  # Compare start point
        self.assertEqual(args[2], (50, 60))  # Compare end point
        self.assertEqual(args[3], (0, 255, 0))  # Compare color
        self.assertEqual(args[4], 2)  # Compare line thickness


class TestOptimizedPath(unittest.TestCase):

    def test_optimized_path_with_mock_distance(self):
        coords = [(0, 0), (1, 2), (3, 4), (5, 6)]
        start = (0, 0)

        # Use patch to mock the behavior of distance function
        with patch(
            "modules.image.select_regions.distance", side_effect=cycle([1])
        ) as mock_distance:
            result = optimized_path(coords, start)

            # Add debug output
            print("Distance call count:", mock_distance.call_count)

            # Expected call count is 6 times
            self.assertEqual(mock_distance.call_count, 6)

    def test_optimized_path_without_start(self):
        coords = [(0, 0), (1, 2), (3, 4), (5, 6)]

        # Verify when start is not provided, it defaults to the first point
        result = optimized_path(coords)
        expected_result = np.array([(0, 0), (1, 2), (3, 4), (5, 6)])
        np.testing.assert_array_equal(result, expected_result)

    def test_optimized_path_with_single_point(self):
        coords = [(0, 0)]

        # Handle the case of a single point, the result should return the input single point
        result = optimized_path(coords)
        expected_result = np.array([(0, 0)])
        np.testing.assert_array_equal(result, expected_result)


if __name__ == "__main__":
    unittest.main()



from modules.ift.younglaplace.younglaplace import YoungLaplaceFitResult
from modules.ift.circle import circle_fit, CircleFitResult
from modules.ift.needle import needle_fit, NeedleFitResult
from modules.image.select_regions import get_ift_regions
from utils.geometry import Rect2, Vector2
from utils.misc import rotation_mat2d

from typing import NamedTuple, Optional, Tuple, List
import math
import cv2
import numpy as np

__all__ = ("PendantFeatures", "extract_pendant_features", "find_pendant_apex")

# Math constants.
PI = math.pi

RotatedRect = Tuple[Vector2[float], Vector2[float], Vector2[float], Vector2[float]]


class PendantFeatures(NamedTuple):
    labels: np.ndarray

    drop_points: np.ndarray = np.empty((2, 0), dtype=int)

    drop_apex: Optional[Vector2[int]] = None
    drop_radius: Optional[float] = None
    drop_rotation: Optional[float] = None

    needle_rect: Optional[RotatedRect] = None
    needle_diameter: Optional[float] = None

    def __eq__(self, other: "PendantFeatures") -> bool:
        if not isinstance(other, PendantFeatures):
            return False

        for v1, v2 in zip(self, other):
            if isinstance(v1, np.ndarray):
                if not (v1 == v2).all():
                    return False
            else:
                if not (v1 == v2):
                    return False
        else:
            return True


def extract_pendant_features(
    image,
    drop_region: Optional[Rect2[int]] = None,
    needle_region: Optional[Rect2[int]] = None,
    *,
    thresh1: float = 80.0,
    thresh2: float = 160.0,
    labels: bool = False,
) -> PendantFeatures:
    """
    Extract needle and drop features from the given image.
    If the regions are not provided, they will be automatically detected.
    Returns a PendantFeatures object containing the extracted features.
    """

    if drop_region is None or needle_region is None:
        automated_drop_region, automated_needle_region = get_ift_regions(image)

        if drop_region is None:
            drop_region = automated_drop_region
        if needle_region is None:
            needle_region = automated_needle_region

    if drop_region is not None:
        drop_image = image[
            drop_region.y0 : drop_region.y1 + 1, drop_region.x0 : drop_region.x1 + 1
        ]
        print(
            f"Drop region: {drop_region.x0}, {drop_region.y0}, {drop_region.x1}, {drop_region.y1}"
        )
        print(f"Drop image shape: {drop_image.shape}")

    if needle_region is not None:
        needle_image = image[
            needle_region.y0 : needle_region.y1 + 1,
            needle_region.x0 : needle_region.x1 + 1,
        ]

    drop_points = np.empty((2, 0), dtype=int)
    drop_apex = None
    drop_radius = None
    drop_rotation = None

    if drop_image is not None:
        if len(drop_image.shape) > 2:
            drop_image = cv2.cvtColor(drop_image, cv2.COLOR_RGB2GRAY)

        drop_points = _extract_drop_edge(drop_image, thresh1, thresh2)

        # There shouldn't be more points than the perimeter of the image.
        if drop_points.shape[1] < 2 * (image.shape[0] + image.shape[1]):
            ans = find_pendant_apex(drop_points)
            if ans is not None:
                drop_apex, drop_radius, drop_rotation = ans

    needle_points = np.empty((2, 0), dtype=int)
    needle_rect = None
    needle_diameter_px = None

    if needle_image is not None:
        if len(needle_image.shape) > 2:
            needle_image = cv2.cvtColor(needle_image, cv2.COLOR_RGB2GRAY)

        blur = cv2.GaussianBlur(needle_image, ksize=(5, 5), sigmaX=0)
        dx = cv2.Scharr(blur, cv2.CV_16S, dx=1, dy=0)
        dy = cv2.Scharr(blur, cv2.CV_16S, dx=0, dy=1)

        # Use magnitude of gradient squared to get sharper edges.
        mask = dx.astype(float) ** 2 + dy.astype(float) ** 2
        mask = (mask / mask.max() * (2**8 - 1)).astype(np.uint8)
        cv2.adaptiveThreshold(
            mask,
            maxValue=1,
            adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
            thresholdType=cv2.THRESH_BINARY,
            blockSize=5,
            C=0,
            dst=mask,
        )

        # Hack: Thin edges using cv2.Canny()
        needle_edges = cv2.Canny(
            dx=mask * dx, dy=mask * dy, threshold1=0.0, threshold2=0.0
        )

        needle_points = np.array(needle_edges.nonzero()[::-1])

        # Use left and right-most points only for fitting.
        needle_outer_points = np.block(
            [
                [
                    np.argmax(needle_edges, axis=1),
                    (needle_edges.shape[1] - 1)
                    - np.argmax(needle_edges[:, ::-1], axis=1),
                ],
                [np.arange(needle_edges.shape[0]), np.arange(needle_edges.shape[0])],
            ]
        )

        needle_fit_result: NeedleFitResult = needle_fit(needle_outer_points)
        if needle_fit_result is not None:
            needle_residuals = np.abs(needle_fit_result.residuals)
            needle_lmask = needle_fit_result.lmask
            needle_rmask = ~needle_lmask
            needle_lpoints = needle_outer_points[
                :, (needle_residuals < 1.0) & needle_lmask
            ]
            needle_rpoints = needle_outer_points[
                :, (needle_residuals < 1.0) & needle_rmask
            ]
            n_lpoints = needle_lpoints.shape[1]
            n_rpoints = needle_rpoints.shape[1]

            # Make sure there's an even number of points on the left and right sides, otherwise probably a bad
            # fit.
            if (
                n_lpoints > 0
                and n_rpoints > 0
                and abs(n_lpoints - n_rpoints) / (n_lpoints + n_rpoints) < 0.33
            ):
                needle_rho = needle_fit_result.rho
                needle_radius = needle_fit_result.radius
                needle_rotation = needle_fit_result.rotation

                needle_rotation_mat = rotation_mat2d(needle_rotation)
                needle_perp = needle_rotation_mat @ [1, 0]
                needle_lpoints_z = (needle_rotation_mat.T @ needle_lpoints)[1]
                needle_rpoints_z = (needle_rotation_mat.T @ needle_rpoints)[1]
                needle_min_z = min(needle_lpoints_z.min(), needle_rpoints_z.min())
                needle_max_z = max(needle_lpoints_z.max(), needle_rpoints_z.max())
                needle_tip1 = needle_rotation_mat @ [needle_rho, needle_min_z]
                needle_tip2 = needle_rotation_mat @ [needle_rho, needle_max_z]

                needle_rect = (
                    Vector2(needle_tip1 - needle_perp * needle_radius),
                    Vector2(needle_tip1 + needle_perp * needle_radius),
                    Vector2(needle_tip2 - needle_perp * needle_radius),
                    Vector2(needle_tip2 + needle_perp * needle_radius),
                )
                needle_diameter_px = 2 * needle_radius

    if drop_region is not None:
        drop_points += np.reshape(drop_region.position, (2, 1))
        if drop_apex is not None:
            drop_apex += drop_region.position

    if needle_region is not None:
        needle_points += np.reshape(needle_region.position, (2, 1))
        if needle_rect is not None:
            needle_rect = (
                needle_region.position + needle_rect[0],
                needle_region.position + needle_rect[1],
                needle_region.position + needle_rect[2],
                needle_region.position + needle_rect[3],
            )

    if labels:
        labels_array = np.zeros(image.shape[:2], dtype=np.uint8)
        labels_array[tuple(drop_points)[::-1]] = 1
        labels_array[tuple(needle_points)[::-1]] = 2
    else:
        labels_array = None

    return (
        drop_points,
        needle_diameter_px,
        drop_region,
        needle_region,
        image,
        drop_image,
        needle_fit_result,
    )


def _extract_drop_edge(gray: np.ndarray, thresh1: float, thresh2: float) -> np.ndarray:
    blur = cv2.GaussianBlur(gray, ksize=(5, 5), sigmaX=0)
    dx = cv2.Scharr(blur, cv2.CV_16S, dx=1, dy=0)
    dy = cv2.Scharr(blur, cv2.CV_16S, dx=0, dy=1)

    # Use magnitude of gradient squared to get sharper edges.
    grad = dx.astype(float) ** 2 + dy.astype(float) ** 2
    grad /= grad.max()
    grad = (grad * (2**8 - 1)).astype(np.uint8)

    cv2.adaptiveThreshold(
        grad,
        maxValue=255,
        adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        thresholdType=cv2.THRESH_BINARY,
        blockSize=5,
        C=0,
        dst=grad,
    )

    # Hack: Use cv2.Canny() to do non-max suppression edge thinning.
    mask = _largest_connected_component(grad)
    edges = cv2.Canny(dx * mask, dy * mask, thresh1, thresh2)
    points = np.array(edges.nonzero()[::-1])

    return points


def _largest_connected_component(gray: np.ndarray) -> np.ndarray:
    mask: np.ndarray

    # Values returned are n_labels, labels, stats, centroids.
    _, labels, stats, _ = cv2.connectedComponentsWithStats(gray, connectivity=4)

    ix = np.argsort(stats[:, cv2.CC_STAT_WIDTH] * stats[:, cv2.CC_STAT_HEIGHT])[::-1]
    if len(ix) > 1:
        if ix[0] == 0:
            # Label 0 is the background.
            biggest_label = ix[1]
        else:
            biggest_label = ix[0]

        mask = labels == biggest_label
    else:
        mask = np.ones(gray.shape, dtype=bool)

    return mask


def find_pendant_apex(data: Tuple[np.ndarray, np.ndarray]) -> Optional[tuple]:

    x, y = data

    if len(x) == 0 or len(y) == 0:
        return None

    xc = x.mean()
    yc = y.mean()
    radius = np.hypot(x - xc, y - yc).mean()

    # Fit a circle to the most circular part of the data.
    circle_fit_result: CircleFitResult = circle_fit(
        data,
        loss="arctan",
        f_scale=radius / 100,
    )
    if circle_fit_result is None:
        return None

    xc, yc = circle_fit_result.center
    radius = circle_fit_result.radius
    resids = np.abs(circle_fit_result.residuals)
    resids_50ptile = np.quantile(resids, 0.5)

    # The somewhat circular-ish part of the drop profile.
    bowl_mask = resids < 10 * resids_50ptile
    bowl_x = x[bowl_mask]
    bowl_y = y[bowl_mask]

    if len(bowl_x) == 0:
        return None

    # Don't need these variables anymore.
    del resids, resids_50ptile

    # Find the symmetry axis of bowl.
    Ixx, Iyy, Ixy = _calculate_inertia(bowl_x, bowl_y)

    # Eigenvector calculation for a symmetric 2x2 matrix.
    rotation = 0.5 * np.arctan2(2 * Ixy, Ixx - Iyy)
    unit_r = np.array([np.cos(rotation), np.sin(rotation)])
    unit_z = np.array([-np.sin(rotation), np.cos(rotation)])

    bowl_r = unit_r @ [bowl_x - xc, bowl_y - yc]
    bowl_z = unit_z @ [bowl_x - xc, bowl_y - yc]
    bowl_r_ix = np.argsort(bowl_r)
    bowl_z_ix = np.argsort(bowl_z)

    # Calculate "asymmetry" along each axis. We define this to be the squared difference between the left and
    # right points, integrated along the axis.
    ma_kernel = np.ones(max(1, len(bowl_r) // 10))
    ma_kernel /= len(ma_kernel)
    asymm_r = (
        np.convolve((bowl_z - bowl_z.mean())[bowl_r_ix], ma_kernel, mode="valid") ** 2
    ).sum()
    asymm_z = (
        np.convolve((bowl_r - bowl_r.mean())[bowl_z_ix], ma_kernel, mode="valid") ** 2
    ).sum()
    if asymm_z > asymm_r:
        # Swap axes so z is the symmetry axis.
        rotation -= PI / 2
        unit_r, unit_z = -unit_z, unit_r
        bowl_r, bowl_z = -bowl_z, bowl_r
        bowl_r_ix, bowl_z_ix = bowl_z_ix[::-1], bowl_r_ix

    # No longer useful variables (and potentially incorrect after axes swapping).
    del asymm_r, asymm_z

    bowl_z_hist, _ = np.histogram(bowl_z, bins=2 + len(bowl_z) // 10)
    if bowl_z_hist.argmax() > len(bowl_z_hist) / 2:
        # Rotate by 180 degrees since points are accumulating (where dz/ds ~ 0) at high z, i.e. drop apex is
        # not on the bottom.
        rotation += PI
        unit_r *= -1
        unit_z *= -1
        bowl_r *= -1
        bowl_z *= -1
        bowl_r_ix = bowl_r_ix[::-1]
        bowl_z_ix = bowl_z_ix[::-1]

    bowl_z_ix_apex_arc_stop = np.searchsorted(
        np.abs(bowl_r), 0.3 * radius, side="right", sorter=bowl_z_ix
    )
    apex_arc_ix = bowl_z_ix[:bowl_z_ix_apex_arc_stop]
    apex_arc_x = bowl_x[apex_arc_ix]
    apex_arc_y = bowl_y[apex_arc_ix]

    if len(apex_arc_ix) > 10:
        # Fit another circle to a smaller arc around the apex. Points within 0.3 radians of the apex should
        # have roughly constant curvature across typical Bond values.
        circle_fit_result = circle_fit(
            np.array([apex_arc_x, apex_arc_y]),
            xc=xc,
            yc=yc,
        )
        if circle_fit_result is not None:
            xc, yc = circle_fit_result.center
            radius = circle_fit_result.radius

    apex_x, apex_y = [xc, yc] - radius * unit_z

    # Restrict rotation to [-pi, pi].
    rotation = (rotation + PI) % (2 * PI) - PI

    return Vector2(apex_x, apex_y), radius, rotation


def _calculate_inertia(x: np.ndarray, y: np.ndarray) -> Tuple[float, float, float]:
    x = x - x.mean()
    y = y - y.mean()

    Ixx = (y**2).sum()
    Iyy = (x**2).sum()
    Ixy = -(x @ y)

    return Ixx, Iyy, Ixy


def _circle_residues(params, x, y):
    xc, yc, radius = params
    r = np.hypot(x - xc, y - yc)
    return r - radius


def _circle_jac(params, x, y):
    jac = np.empty((len(x), 3))

    xc, yc, radius = params
    dist_x = x - xc
    dist_y = y - yc
    r = np.hypot(dist_x, dist_y)

    jac[:, 0] = -dist_x / r
    jac[:, 1] = -dist_y / r
    jac[:, 2] = -1

    return jac


def analyze_ift(
    fit_result: YoungLaplaceFitResult,
    *,
    drop_density: float,
    continuous_density: float = 0,
    needle_diameter_mm: float,
    needle_diameter_px: float = 0,
):
    gravity = 9.81
    # 1) figure out how many pixels per millimetre
    px_per_mm = needle_diameter_mm / needle_diameter_px

    # 2) unpack the pixel‐space results
    radius_px = fit_result.radius  # ensure positive
    volume_px = fit_result.volume
    surface_area_px = fit_result.surface_area
    bond = fit_result.bond
    delta_density = abs(drop_density - continuous_density)
    radius = radius_px * px_per_mm
    surface_area = surface_area_px * px_per_mm**2
    volume = volume_px * px_per_mm**3
    IFT = (delta_density * gravity * radius**2 / bond) / 1000

    # 5) Worthington number
    worthington = (
        (delta_density * gravity * volume) / (PI * IFT * needle_diameter_mm)
    ) / 1000

    return [IFT, volume, surface_area, bond, worthington, None]


def auto_crop(
    img: np.ndarray,
    padding: int = 5,
    area_thresh: int = 1,
    drop_frac: float = 1,
    scharr_block: int = 5,
    canny1: float = 80,
    canny2: float = 160,
) -> Tuple[np.ndarray, tuple]:

    original = img.copy()
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    H, W = gray.shape
    # Applying 7x7 Gaussian Blur
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    # 1) rough mask via inverted Otsu + largest CC
    threshold = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]
    # n_labels, labels, stats, _ = cv2.connectedComponentsWithStats(thr, 8)

    analysis = cv2.connectedComponentsWithStats(threshold, 8)
    (totalLabels, label_ids, stats, centroids) = analysis
    if totalLabels <= 1:
        print("drops found.")
        return original, None

        # Filter for components entering from the top and find the largest among them
    top_entering_blobs_indices = []
    for i in range(1, totalLabels):  # Iterate through all components (0 is background)
        if stats[i, cv2.CC_STAT_TOP] == 0:
            top_entering_blobs_indices.append(i)

    if not top_entering_blobs_indices:
        print("No blobs found entering from the top of the frame.")
        return original, None

    # Select the largest blob among those entering from the top
    blob = -1
    max_area_found = -1
    for idx in top_entering_blobs_indices:
        area = stats[idx, cv2.CC_STAT_AREA]
        if area > max_area_found:
            max_area_found = area
            blob = idx

    if blob == -1:  # Should not happen if top_entering_blobs_indices was populated
        print("Error selecting largest top-entering blob.")
        return original, None

    if stats[blob, cv2.CC_STAT_AREA] < area_thresh:
        print(
            f"Largest top-entering blob (ID: {blob}, Area: {stats[blob, cv2.CC_STAT_AREA]}) too small (threshold: {area_thresh})."
        )
        return original, None

    x, y, bw, bh, _ = stats[blob]
    # print(f"Blob ID: {blob}, Area: {stats[blob, cv2.CC_STAT_AREA]}, Bounding box: (x= {x}, y= {y}) - (w= {bw}, h= {bh})")
    roi = gray[y : y + bh, x : x + bw]
    roi_mask = (label_ids[y : y + bh, x : x + bw] == blob).astype(np.uint8) * 255

    # 2) Scharr→adaptiveThresh→Canny inside ROI
    blur = cv2.GaussianBlur(roi, (scharr_block, scharr_block), 0)
    dx = cv2.Scharr(blur, cv2.CV_16S, 1, 0)
    dy = cv2.Scharr(blur, cv2.CV_16S, 0, 1)
    grad = dx.astype(float) ** 2 + dy.astype(float) ** 2
    grad = np.uint8(255 * (grad / grad.max()))
    cv2.adaptiveThreshold(
        grad,
        255,
        cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        cv2.THRESH_BINARY,
        blockSize=scharr_block,
        C=0,
        dst=grad,
    )
    grad = cv2.bitwise_and(grad, grad, mask=roi_mask)
    edges = cv2.Canny(grad, canny1, canny2)

    # 3) collect edge points *only* in the bottom (1-drop_frac) of ROI:
    pts = np.column_stack(np.nonzero(edges))  # (row, col)
    if len(pts) < 20:
        print("Too few edges.")
        return original, None
    # Calculate width of the blob at each row within the ROI
    blob_widths_at_each_row = np.sum(roi_mask // 255, axis=1)
    # Define the upper part of the blob to estimate needle width
    # Consider top 15% of blob height, but at least 3px and at most 20px
    initial_needle_scan_height = max(3, min(int(bh * 0.15), 500))
    # Ensure scan height is less than blob height for sensible indexing
    if initial_needle_scan_height >= bh and bh > 0:
        initial_needle_scan_height = bh - 1
    elif bh == 0:  # Should not happen if blob found
        initial_needle_scan_height = 0
    avg_needle_width = 0
    if initial_needle_scan_height > 0:
        candidate_needle_widths = blob_widths_at_each_row[:initial_needle_scan_height]
        if np.any(candidate_needle_widths > 0):
            avg_needle_width = np.median(
                candidate_needle_widths[candidate_needle_widths > 0]
            )

    if (
        avg_needle_width <= 1
    ):  # If needle width is too small or not found, use a small portion of top widths
        fallback_scan_height = max(1, int(bh * 0.05))
        if bh > 0 and np.any(blob_widths_at_each_row[:fallback_scan_height] > 0):
            avg_needle_width = np.mean(
                blob_widths_at_each_row[:fallback_scan_height][
                    blob_widths_at_each_row[:fallback_scan_height] > 0
                ]
            )
        if avg_needle_width <= 0:  # Absolute fallback if still no valid width
            avg_needle_width = 1
    expansion_factor = 1  # Drop should be significantly wider than needle
    search_start_row = initial_needle_scan_height

    found_expansion_point = False
    calculated_row_thresh = 0
    for current_row in range(search_start_row, bh):
        # Add small absolute diff
        if (
            blob_widths_at_each_row[current_row] > expansion_factor * avg_needle_width
            and blob_widths_at_each_row[current_row] > avg_needle_width + 2
        ):
            calculated_row_thresh = current_row
            found_expansion_point = True
            # print(f"Dynamic row_thresh: Initial needle width ~{avg_needle_width:.1f}px. Expansion at row {current_row} (width {blob_widths_at_each_row[current_row]}px).")
            break

    if not found_expansion_point:
        print(
            f"No clear expansion point found. Initial needle width ~{avg_needle_width:.1f}px. Using drop_frac based fallback: {calculated_row_thresh}."
        )
        # Already defaulted to drop_frac based
    row_thresh = calculated_row_thresh

    # print(f"Row threshold for drop detection: {row_thresh} (calculated)")
    # --- End of dynamic row_thresh determination ---
    def create_bounds_array(x1, y1, x2, y2):
        # Ensure coordinates are integers for array creation if they are not None
        if any(v is None for v in [x1, y1, x2, y2]):
            return None
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        return np.array(
            [
                [[x1, y1, x2, y1]],  # top edge
                [[x2, y1, x2, y2]],  # right edge
                [[x2, y2, x1, y2]],  # bottom edge
                [[x1, y2, x1, y1]],  # left edge
            ],
            dtype=np.int32,
        )

    # Create the bounding box for the entire ROI
    x1_roi = x
    y1_roi = y
    x2_roi = x + bw
    y2_roi = y + bh
    # Create the absolute bounding box for the entire ROI
    abs_bounds = create_bounds_array(x1_roi, y1_roi, x2_roi, y2_roi)
    # Create the cropped bounding box for the entire ROI
    crop_bounds = create_bounds_array(x1_roi, y1_roi, x2_roi, y2_roi)
    crop_roi = original[y1_roi:y2_roi, x1_roi:x2_roi]
    if not found_expansion_point:
        needle_pts = pts[pts[:, 0] < row_thresh]
        pts_xy_needle = needle_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_needle.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the needle points, *relative to the ROI's top-left corner*
            needle_x1_roi = np.min(pts_xy_needle[:, 0])
            needle_x2_roi = np.max(pts_xy_needle[:, 0])
            needle_y1_roi = np.min(pts_xy_needle[:, 1])
            needle_y2_roi = np.max(pts_xy_needle[:, 1])
            needle_abs_bounds = create_bounds_array(
                x + needle_x1_roi - padding,
                y + needle_y1_roi,
                x + needle_x2_roi + padding,
                y + needle_y2_roi - padding,
            )
            needle_crop_bounds = create_bounds_array(
                needle_x1_roi - padding,
                needle_y1_roi,
                needle_x2_roi + padding,
                needle_y2_roi - padding,
            )
            # Crop the original image using the final absolute bounding box coordinates
            crop = original[
                needle_y1_roi:needle_y2_roi, x + needle_x1_roi : x + needle_x2_roi
            ]
            needle_rect = Rect2(
                (x + needle_x1_roi, y + needle_y1_roi),
                (x + needle_x2_roi, y + needle_y2_roi),
            )
            return needle_rect
    else:
        needle_pts = pts[pts[:, 0] < row_thresh]
        pts_xy_needle = needle_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_needle.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the needle points, *relative to the ROI's top-left corner*
            needle_x1_roi = np.min(pts_xy_needle[:, 0]) - padding
            needle_x2_roi = np.max(pts_xy_needle[:, 0]) + padding
            needle_y1_roi = np.min(pts_xy_needle[:, 1])
            needle_y2_roi = np.max(pts_xy_needle[:, 1]) - padding
            needle_abs_bounds = create_bounds_array(
                x + needle_x1_roi,
                y + needle_y1_roi,
                x + needle_x2_roi,
                y + needle_y2_roi,
            )
            needle_crop_bounds = create_bounds_array(
                needle_x1_roi - padding, needle_y1_roi, needle_x2_roi, needle_y2_roi
            )
            needle_rect = Rect2(
                (x + needle_x1_roi, y + needle_y1_roi),
                (x + needle_x2_roi, y + needle_y2_roi),
            )

        drop_pts = pts[pts[:, 0] >= row_thresh]
        pts_xy_drop = drop_pts[:, ::-1]  # (x, y) coordinates
        if pts_xy_drop.shape[0] > 0:  # Ensure there are points to process
            # These are min/max coordinates of the drop points, *relative to the ROI's top-left corner*
            drop_x1_roi = np.min(pts_xy_drop[:, 0]) - padding
            drop_x2_roi = np.max(pts_xy_drop[:, 0]) + padding
            drop_y1_roi = np.min(pts_xy_drop[:, 1])  # - padding
            drop_y2_roi = np.max(pts_xy_drop[:, 1]) + padding
            drop_abs_bounds = create_bounds_array(
                x + drop_x1_roi, y + drop_y1_roi, x + drop_x2_roi, y + drop_y2_roi
            )
            drop_crop_bounds = create_bounds_array(
                drop_x1_roi, drop_y1_roi, drop_x2_roi, drop_y2_roi
            )
            drop_rect = Rect2(
                (x + drop_x1_roi, y + drop_y1_roi), (x + drop_x2_roi, y + drop_y2_roi)
            )

        crop = original[drop_y1_roi:drop_y2_roi, x + drop_x1_roi : x + drop_x2_roi]
        return drop_rect, needle_rect



from typing import Sequence
import numpy as np


class YoungLaplaceShape:
    def __init__(self, bond: float) -> None: ...
    def __call__(self, s: float) -> np.ndarray: ...
    def DBo(self, s: float) -> np.ndarray: ...
    def z_inv(self, s: float) -> float: ...
    def closest(self, v: Sequence[float]) -> float: ...
    def volume(self, s: float) -> float: ...
    def surface_area(self, s: float) -> float: ...
    @property
    def bond(self) -> float: ...



"""This code serves as a discrete instance of the elipse fit method of
contact angle analysis.

This is base on the circular fit code taken from the most recent version
of conan - conan-ML_cv1.1/modules/select_regions.py"""

from utils.config import CV2_VERSION

from sklearn.cluster import OPTICS  # for clustering algorithm
from skimage.measure import EllipseModel
from matplotlib.patches import Ellipse
import numpy as np
import matplotlib.pyplot as plt
import cv2
import math
import time


def cluster_optics(sample, out_style="coords", xi=None, eps=None, verbose=0):
    """Takes an array (or list) of the form [[x1,y1],[x2,y2],...,[xn,yn]].
    Clusters are outputted in the form of a dictionary.

    If out_style='coords' each dictionary entry is a group, and points are outputted in coordinate form.
    If out_xy='xy' there are two dictionary entries for each group, one labeled as nx and one as ny
    (where n is the label of the group)

    If xi (float between 0 and 1) is not None and eps is None, then the xi clustering method is used.
    The optics algorithm defines clusters based on the minimum steepness on the reachability plot.
    For example, an upwards point in the reachability plot is defined by the ratio from one point to
    its successor being at most 1-xi.

    If eps (float) is not None and xi is None, then the dbscan clustering method is used. Where eps is the
    maximum distance between two samples for one to be considered as in the neighborhood of the other.

    https://stackoverflow.com/questions/47974874/algorithm-for-grouping-points-in-given-distance
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.OPTICS.html
    """
    if eps != None and xi == None:
        clustering = OPTICS(min_samples=2, cluster_method="dbscan", eps=eps).fit(
            sample
        )  # cluster_method changed to dbscan (so eps can be set)
    elif xi != None and eps == None:
        # original had xi = 0.05, xi as 0.1 in function input
        clustering = OPTICS(min_samples=2, xi=xi).fit(sample)
    else:
        raise ValueError(
            "only one of eps and xi can be chosen but not neither nor both"
        )
    groups = list(set(clustering.labels_))

    if verbose == 2:
        print(clustering.labels_)
    elif verbose == 1:
        print(groups)
    elif verbose == 0:
        pass

    dic = {}
    for n in groups:
        if n not in dic:
            dic[n] = []
        for i in range(len(sample)):
            if clustering.labels_[i] == n:
                dic[n].append(sample[i])

    # separate points and graph
    dic2 = {}
    for k in dic.keys():
        x = []
        y = []
        for i in range(len(dic[k])):
            x.append(dic[k][i][0])
        dic2[str(k) + "x"] = x
        for i in range(len(dic[k])):
            y.append(dic[k][i][1])
        dic2[str(k) + "y"] = y

    if out_style == "coords":
        return dic
    elif out_style == "xy":
        return dic2


def distance1(P1, P2):
    """This function computes the distance between 2 points defined by
    P1 = (x1,y1) and P2 = (x2,y2)"""
    return ((P1[0] - P2[0]) ** 2 + (P1[1] - P2[1]) ** 2) ** 0.5


def optimized_path(coords, start=None):
    """This function finds the nearest point to a point
    coords should be a list in this format coords = [ [x1, y1], [x2, y2] , ...]
    https://stackoverflow.com/questions/45829155/sort-points-in-order-to-have-a-continuous-curve-using-python
    """
    if isinstance(coords, list) == False:
        coords = coords.tolist()
    if 0:
        if isinstance(start, list) == False:
            try:
                start = start.tolist()
            except:
                start = list(start)
    if start is None:
        start = coords[0]
    pass_by = coords
    path = [start]
    pass_by.remove(start)
    while pass_by:
        nearest = min(pass_by, key=lambda x: distance1(path[-1], x))
        path.append(nearest)
        pass_by.remove(nearest)
    path = np.array(path)

    # if there are any large jumps in distance, there is likely a mistake
    # therefore, the points after this jump should be ignored
    if 1:
        dists = []
        for i, point in enumerate(path):
            if i < len(path) - 1:
                dists.append(distance1(path[i], path[i + 1]))
        jump_idx = []
        for i, dist in enumerate(dists):
            if dist > 5:
                jump_idx.append(i)
        if len(jump_idx) > 0:
            path = path[: jump_idx[0]]

    return path


def prepare_hydrophobic(coords, xi=0.8, cluster=False, display=False):
    """takes an array (n,2) of coordinate points, and returns the left and right halfdrops of the contour.
    xi determines the minimum steepness on the reachability plot that constitutes a cluster boundary of the
    clustering algorithm.
    deg is the degree of the polynomial used to describe the shape of the droplet.

    This code is adapted from the prepare module, but this version differs in that it assumes that the drop
    is hydrophobic."""
    coords = coords.astype(float)
    # scan for clusers to remove noise and circle from lensing effect
    ################  MAY NEED TO OPTIMIZE eps/xi TO FIND APPROPRIATE GROUPINGS  ####################
    if cluster:  # turn this off bc using synthetic drops without lensing effect
        input_contour = coords
        dic, dic2 = cluster_optics(input_contour, xi=xi), cluster_optics(
            input_contour, out_style="xy", xi=xi
        )

        # print("number of groups: ",len(list(dic.keys())))

        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(list(dic.keys())))))
        for k in dic.keys():
            plt.plot(dic2[str(k) + "x"], dic2[str(k) + "y"], "o", color=next(colors))
        plt.title(str(len(dic.keys())) + " groups found by clustering")
        plt.show()
        plt.close()
        maxkey = max(dic, key=lambda k: len(dic[k]))

        # print('key to longest dictionary entry is: ',maxkey)

        # take the longest group
        longest = dic[maxkey]

    # flip contour so that min and max values are correct
    for coord in coords:
        coord[1] = -coord[1]

    longest = coords

    # print("first few coordinates of the longest contour: ",longest[:3])

    xlongest = []
    ylongest = []
    for i in range(len(longest)):
        xlongest.append(longest[i][0])
        ylongest.append(longest[i][1])

    # print("first few x coordinates of the longest contour: ",xlongest[:3])
    # print("first few y coordinates of the longest contour: ",ylongest[:3])

    # Find a appropriate epsilon value for cluster_optics, this will remove noise in the bottom 10% of the drop
    # .   most importantly noise is reduced at contact points.

    # variables in this process are how much and what part of the top of the droplet we use to be representative of
    # the full contour, and whether we use the max(distance) between points or the average between points, or
    # a scalar value of either.

    xtop = []  # isolate top 90% of drop
    ytop = []
    percent = 0.3
    # print('Isolate the top ',100-(percent*100),'% of the contour:')
    for n, y in enumerate(ylongest):
        if y > min(ylongest) + (max(ylongest) - min(ylongest)) * percent:
            xtop.append(xlongest[n])
            ytop.append(y)
    xtop = np.array(xtop)
    ytop = np.array(ytop)

    top = []
    for n, x in enumerate(xtop):
        top.append([xtop[n], ytop[n]])
    top = np.array(top)
    top_array = optimized_path(top)

    dists = []  # find the average distance between consecutive points
    for n, co in enumerate(top_array):
        if 1 < n:
            a = top_array[n]
            dist = np.linalg.norm(top_array[n] - top_array[n - 1])
            dists.append(dist)

    if display:
        # print(dists)
        print()
        print("Max dist between points is: ", max(dists))
        print("Average dist between points is: ", sum(dists) / len(dists))
        print("20% over the Max dist is: ", max(dists) * 1.2)
        print()
        print("Sort using cluster_optics with an epsilon value of ", max(dists) * 1.2)

    # how epsilon is chosen here is important
    # eps = (sum(dists)/len(dists))*2 # eps is 2 times the average distance between points
    # eps is 2.5 times the average distance between points
    eps = (sum(dists) / len(dists)) * 1.5
    input_contour = longest
    dic, dic2 = cluster_optics(input_contour, eps=eps), cluster_optics(
        input_contour, out_style="xy", eps=eps
    )

    # print("number of groups: ",len(list(dic.keys())))
    if display:
        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(list(dic.keys())))))
        for k in dic.keys():
            plt.plot(dic2[str(k) + "x"], dic2[str(k) + "y"], "o", color=next(colors))
        plt.title("Groups found by clustering with epsilon value of " + str(eps))
        plt.show()
        plt.close()
    maxkey = max(dic, key=lambda k: len(dic[k]))
    longest = dic[maxkey]
    # print("first few coordinates of the longest contour: ",longest[:3])

    xlongest = []
    ylongest = []
    for i in range(len(longest)):
        xlongest.append(longest[i][0])
        ylongest.append(longest[i][1])

    outline = np.empty((len(xlongest), 2))
    for i in range(len(xlongest)):
        outline[i, [0]] = xlongest[i]
        outline[i, [1]] = ylongest[i]

    ############################

    # find the apex of the drop and split the contour into left and right sides

    xtop = []  # isolate top 90% of drop
    ytop = []
    # percent = 0.1 # already defined
    # print('isolate the top ',100-(percent*100),'% of the contour:')
    for n, y in enumerate(ylongest):
        if y > min(ylongest) + (max(ylongest) - min(ylongest)) * percent:
            xtop.append(xlongest[n])
            ytop.append(y)
    xapex = (max(xtop) + min(xtop)) / 2

    l_drop = []
    r_drop = []
    for n in longest:
        if n[0] <= xapex:
            l_drop.append(n)
        if n[0] >= xapex:
            r_drop.append(n)
    l_drop = np.array(l_drop)
    r_drop = np.array(r_drop)

    # transpose both half drops so that they both face right and the apex of both is at 0,0
    r_drop[:, 0] = r_drop[:, 0] - xapex
    l_drop[:, 0] = -l_drop[:, 0] + xapex

    if display:
        plt.plot(r_drop[:, [0]], r_drop[:, [1]], "b,")
        # plt.show()
        # plt.close()
        plt.plot(l_drop[:, [0]], l_drop[:, [1]], "r,")
        # plt.gca().set_aspect('equal', adjustable='box')
        # plt.xlim([470,530])
        # plt.ylim([-188,-190])
        plt.show()
        plt.close()

    #############################

    # the drop has been split in half

    # this system has a user input which gives a rough indication of the contact point and the surface line

    # isolate the bottom 5% of the contour near the contact point

    drops = {}
    counter = 0
    crop_drop = {}
    CPs = {}
    for halfdrop in [l_drop, r_drop]:
        new_halfdrop = sorted(
            halfdrop.tolist(), key=lambda x: (x[0], -x[1])
        )  # top left to bottom right
        new_halfdrop = optimized_path(new_halfdrop)  # [::-1]

        xnew_halfdrop = new_halfdrop[:, [0]].reshape(len(new_halfdrop[:, [0]]))
        ynew_halfdrop = new_halfdrop[:, [1]].reshape(len(new_halfdrop[:, [1]]))

        # isolate the bottom of the drop to help identify contact points (may not need to do this for all scenarios)
        bottom = []
        top = []  # will need this later
        # print('isolate the bottom ',percent*100,'% of the contour:') # percent defined above
        div_line_value = (
            min(new_halfdrop[:, [1]])
            + (max(new_halfdrop[:, [1]]) - min(new_halfdrop[:, [1]])) * percent
        )
        for n in new_halfdrop:
            if n[1] < div_line_value:
                bottom.append(n)
            else:
                top.append(n)

        bottom = np.array(bottom)
        top = np.array(top)

        xbottom = bottom[:, [0]].reshape(len(bottom[:, [0]]))
        ybottom = bottom[:, [1]].reshape(len(bottom[:, [1]]))
        xtop = top[:, [0]].reshape(len(top[:, [0]]))
        ytop = top[:, [1]].reshape(len(top[:, [1]]))

        # print('max x value of halfdrop is: ',max(xhalfdrop))

        if 1:  # plot the bottom 10% of the contour
            plt.plot(xbottom, ybottom, "b,")
            plt.title("bottom 10% of the contour")
            # plt.xlim([130,200])
            plt.show()
            plt.close()

        #### Continue here assuming that the drop is hydrophobic ####
        if 1:
            # order all halfdrop points using optimized_path (very quick but occasionally makes mistakes)

            xCP = min(xbottom)
            # yCP = min([coord[1] for coord in new_halfdrop if coord[0]==xCP])
            yCP = max([coord[1] for coord in bottom if coord[0] == xCP])
            CPs[counter] = [xCP, yCP]

            if display:  # check
                plt.plot(new_halfdrop[:, 0], new_halfdrop[:, 1])
                plt.show()
                plt.close()

            # remove surface line past the contact point
            index = new_halfdrop.tolist().index(CPs[counter])  # ?

            new_halfdrop = new_halfdrop[: index + 1]

            if 0:
                xCP_index = [i for i, j in enumerate(xnew_halfdrop) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(ynew_halfdrop) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                new_halfdrop = np.zeros((len(xnew_halfdrop), 2))
                for n in range(len(xnew_halfdrop)):
                    new_halfdrop[n, [0]] = xnew_halfdrop[n]
                    new_halfdrop[n, [1]] = ynew_halfdrop[n]
                # print('first 3 points of new_halfdrop are: ',new_halfdrop[:3])
                # print('length of new_halfdrop is: ',len(new_halfdrop))

                if xCP_index == yCP_index:
                    if (
                        new_halfdrop[xCP_index[0] + 1][1]
                        > new_halfdrop[xCP_index[0] - 1][1]
                    ):
                        new_halfdrop = new_halfdrop[xCP_index[0] :]
                    else:
                        new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                else:
                    raise_error = True
                    for x in xCP_index:
                        for y in yCP_index:
                            if x == y:
                                raise_error = False
                                xCP_index = [x]
                                yCP_index = [y]
                                # print('indexes of the CP are: ',xCP_index[0],', ',yCP_index[0])
                                if (
                                    new_halfdrop[xCP_index[0] + 1][1]
                                    > new_halfdrop[xCP_index[0] - 1][1]
                                ):
                                    new_halfdrop = new_halfdrop[xCP_index[0] :]
                                else:
                                    new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                    if raise_error == True:
                        print(
                            "The index of the contact point x value is: ",
                            new_halfdrop[xCP_index],
                        )
                        print(
                            "The index of the contact point y value is: ",
                            new_halfdrop[yCP_index],
                        )
                        raise "indexes of x and y values of the contact point are not the same"

        if 0:
            # order all halfdrop points using two-opt (the slower method)

            # before any ordering is done, chop off the surface line that is past the drop edge
            del_indexes = []
            for index, coord in enumerate(bottom):
                if coord[0] > max(xtop):
                    del_indexes.append(index)
                if coord[1] < yCP:
                    del_indexes.append(index)
            # halfdrop = np.delete(halfdrop,del_indexes)
            xbot = np.delete(bottom[:, 0], del_indexes)
            ybot = np.delete(bottom[:, 1], del_indexes)

            # bottom = np.empty((len(xbot),2))
            # for n, coord in enumerate(bottom):
            #    bottom[n,0] = xbot[n]
            #    bottom[n,1] = ybot[n]

            bottom = np.array(list(zip(xbot, ybot)))

            # print('shape of another_halfdrop is: '+ str(type(another_halfdrop)))
            print("first few points of halfdrop are: ", halfdrop[:3])
            if display:
                colors = iter(jet(np.linspace(0, 1, len(bottom))))
                for n, coord in enumerate(bottom):
                    plt.plot(bottom[n, 0], bottom[n, 1], "o", color=next(colors))
                plt.title("Cropped bottom of hafldrop")
                plt.show()
                plt.close()

            # order points using traveling salesman two_opt code
            bottom = bottom[::-1]  # start at the top
            print(
                "Starting first coordinate of bottom slice of halfdrop is: ", bottom[0]
            )
            # increase improvement_threshold from 0.1 to 0.01
            new_bot, _ = two_opt(bottom, 0.01)
            if new_bot[0, 1] < new_bot[-1, 1]:
                new_bot = new_bot[::-1]
            xbot, ybot = new_bot[:, [0]], new_bot[:, [1]]

            if display:  # display
                colors = iter(jet(np.linspace(0, 1, len(bottom))))
                for n, coord in enumerate(bottom):
                    plt.plot(new_bot[n, 0], new_bot[n, 1], "o", color=next(colors))
                plt.title("Cropped ordered new_bot of halfdrop (starting from blue)")
                plt.show()
                plt.close()

                plt.plot(new_bot[:, 0], new_bot[:, 1])
                plt.title("Cropped ordered new_bot of halfdrop (line)")
                plt.show()
                plt.close()

            # order the top 90% so that the y value decreases

            print(
                "Sorting top ", 100 - (percent * 100), "% of the contour by y value..."
            )
            new_top = sorted(list(top), key=lambda x: x[1], reverse=True)
            new_top = np.array(new_top)

            # remove surface by deleting points after the contact point
            xCP_indexs = [i for i, j in enumerate(xbot) if j == xCP]
            # print('xCP_index is: ',xCP_index)
            yCP_indexs = [i for i, j in enumerate(ybot) if j == yCP]
            # print('yCP_index is: ',yCP_index)

            for xCP_index in xCP_indexs:
                for yCP_index in yCP_indexs:
                    if xCP_index == yCP_index:
                        try:
                            if ybot[yCP_index + 2] > ybot[yCP_index - 1]:
                                new_bot = np.zeros((len(xbot[yCP_index:]), 2))
                                for n in range(len(new_bot)):
                                    new_bot[n, [0]] = xbot[xCP_index + n]
                                    new_bot[n, [1]] = ybot[yCP_index + n]
                            else:
                                new_bot = np.zeros((len(xbot[:yCP_index]), 2))
                                for n in range(len(new_bot)):
                                    new_bot[n, [0]] = xbot[n]
                                    new_bot[n, [1]] = ybot[n]
                        except:
                            try:
                                if ybot[yCP_index] > ybot[yCP_index - 2]:
                                    new_bot = np.zeros((len(xbot[yCP_index:]), 2))
                                    for n in range(len(new_bot)):
                                        new_bot[n, [0]] = xbot[xCP_index + n]
                                        new_bot[n, [1]] = ybot[yCP_index + n]
                                else:
                                    new_bot = np.zeros((len(xbot[:yCP_index]), 2))
                                    for n in range(len(new_bot)):
                                        new_bot[n, [0]] = xbot[n]
                                        new_bot[n, [1]] = ybot[n]
                            except:
                                print("xCP_indexs are: ", xCP_indexs)
                                print("yCP_indexs are: ", yCP_indexs)
                                raise "indexes of x and y values of the contact point are not the same"
            new_halfdrop = np.concatenate((new_top, new_bot))

        if 0:  # order the points so that the baseline can be removed
            # before any ordering is done, chop off the surface line that is past the drop edge
            del_indexes = []
            for index, coord in enumerate(halfdrop):
                if coord[0] > max(xtop):
                    del_indexes.append(index)
            # halfdrop = np.delete(halfdrop,del_indexes)
            xhalfdrop = np.delete(xhalfdrop, del_indexes)
            yhalfdrop = np.delete(yhalfdrop, del_indexes)
            # print('shape of another_halfdrop is: '+ str(type(another_halfdrop)))
            # print('first few points of halfdrop are: ',halfdrop[:3])

            # order half contour points
            xx, yy = sort_to_line(xhalfdrop, yhalfdrop)
            add_top = False
            # print('length of halfdrop is: ', len(halfdrop))
            # print('length of xbottom is: ', len(xbottom))

            # if xx[0]<1: # then graph starts at the top
            surface_past_drop_index = []
            for n, x in enumerate(xx):
                if x > max(xtop):
                    surface_past_drop_index.append(n)
                    # xx = xx[:max(xtop)point]
            # print('Indexes of contour points past drop: ',surface_past_drop_index)

            # if the sort method will not work
            # assumes that the error is on the surface somewhere, so uses bottom of contour
            if len(xx) < len(xhalfdrop):
                add_top = True
                print()
                print(
                    "sort_to_line is not utilising the full contour, alternate ordering method being used"
                )
                print("check bottom 10% of contour...")
                # this method is much slower than the above, so use as few points as possible
                bot_list = []
                for n in range(len(xbottom)):
                    if xbottom[n] < max(xtop):
                        bot_list.append([xbottom[n], ybottom[n]])
                bot_array = np.array(bot_list)
                # 119.8 seconds for 247 points
                new_order, _ = two_opt(bot_array)

                xbot, ybot = new_order[:, [0]], new_order[:, [1]]

                if display:
                    plt.plot(xbot, ybot, "b-")
                    plt.title("Bottom of half drop, new order")
                    plt.show()
                    plt.close()

                xCP_index = [i for i, j in enumerate(xbot) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(ybot) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                if 0:
                    new_bot = np.zeros((len(xbot), 2))
                    for n in range(len(xx)):
                        new_bot[n, [0]] = xbot[n]
                        new_bot[n, [1]] = ybot[n]
                # xbot[xCP_index[0]:]
                if xCP_index == yCP_index:
                    if ybot[yCP_index[0] + 1] > ybot[yCP_index[0] - 1]:
                        new_bot = np.zeros((len(xbot[yCP_index[0] :]), 2))
                        for n in range(len(new_bot)):
                            new_bot[n, [0]] = xbot[xCP_index[0] + n]
                            new_bot[n, [1]] = ybot[yCP_index[0] + n]
                    else:
                        new_bot = np.zeros((len(xbot[: yCP_index[0]]), 2))
                        for n in range(len(new_bot)):
                            new_bot[n, [0]] = xbot[n]
                            new_bot[n, [1]] = ybot[n]
                else:
                    raise "indexes of x and y values of the contact point are not the same"

                # combine new_bot with top_array to give the isolated drop contour without surface
                if 0:
                    top_array = np.zeros((len(xtop), 2))
                    for n in range(len(xtop)):
                        top_array[n, [0]] = xtop[n]
                        top_array[n, [1]] = ytop[n]

                new_halfdrop = np.concatenate((top, new_bot))

                # re-order to check that the error was at the surface line
                xx, yy = sort_to_line(new_halfdrop[:, [0]], new_halfdrop[:, [1]])
                if len(xx) < len(
                    new_halfdrop
                ):  # then the error was in the top 90% of the drop
                    print("Checking top 90% of contour...")
                    new_top, _ = two_opt(top)
                    new_halfdrop = np.concatenate((new_top, new_bot))
                    xx, yy = sort_to_line(new_halfdrop[:, [0]], new_halfdrop[:, [1]])

            else:  # if sort_to_line worked as expected
                # find the indexs of the contact point and chop off the ends
                xCP_index = [i for i, j in enumerate(xx) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(yy) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                new_halfdrop = np.zeros((len(xx), 2))
                for n in range(len(xx)):
                    new_halfdrop[n, [0]] = xx[n]
                    new_halfdrop[n, [1]] = yy[n]
                # print('first 3 points of new_halfdrop are: ',new_halfdrop[:3])
                # print('length of new_halfdrop is: ',len(new_halfdrop))

                if xCP_index == yCP_index:
                    if (
                        new_halfdrop[xCP_index[0] + 1][1]
                        > new_halfdrop[xCP_index[0] - 1][1]
                    ):
                        new_halfdrop = new_halfdrop[xCP_index[0] :]
                    else:
                        new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                else:
                    raise_error = True
                    for x in xCP_index:
                        for y in yCP_index:
                            if x == y:
                                raise_error = False
                                xCP_index = [x]
                                yCP_index = [y]
                                # print('indexes of the CP are: ',xCP_index[0],', ',yCP_index[0])
                                if (
                                    new_halfdrop[xCP_index[0] + 1][1]
                                    > new_halfdrop[xCP_index[0] - 1][1]
                                ):
                                    new_halfdrop = new_halfdrop[xCP_index[0] :]
                                else:
                                    new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                    if raise_error == True:
                        print(
                            "The index of the contact point x value is: ",
                            new_halfdrop[xCP_index],
                        )
                        print(
                            "The index of the contact point y value is: ",
                            new_halfdrop[yCP_index],
                        )
                        raise "indexes of x and y values of the contact point are not the same"

        if counter == 0:
            drops[counter] = new_halfdrop[::-1]
        else:
            drops[counter] = new_halfdrop

        if display:  # display
            jet = plt.get_cmap("jet")
            colors = iter(jet(np.linspace(0, 1, len(new_halfdrop))))
            for k in new_halfdrop:
                plt.plot(k[0], k[1], "o", color=next(colors))
            plt.title("outputted halfdrop")
            plt.axis("equal")
            plt.show()
            plt.close()

        counter += 1

    # reflect the left drop and combine left and right

    profile = np.empty((len(drops[0]) + len(drops[1]), 2))
    for i, n in enumerate(drops[0]):
        flipped = n
        flipped[0] = -flipped[0]
        profile[i] = flipped
    for i, n in enumerate(drops[1]):
        profile[len(drops[0]) + i] = n
    CPs[0][0] = -CPs[0][0]

    if display:
        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(profile))))
        for k in profile:
            plt.plot(k[0], k[1], "o", color=next(colors))
        plt.title("final output")
        # plt.plot(profile[:,0],profile[:,1],'b')
        plt.show()
        plt.close()

        plt.title("final output")
        plt.plot(profile[:, 0], profile[:, 1], "b")
        plt.show()
        plt.close()

    # flip upside down again so that contour follows image indexing
    # and transform to the right so that x=0 is no longer in line with apex
    for coord in profile:
        coord[1] = -coord[1]
        coord[0] = coord[0] + xapex
    for n in [0, 1]:
        CPs[n][1] = -CPs[n][1]
        CPs[n][0] = CPs[n][0] + xapex

    # flip original contour back to original orientation
    for coord in coords:
        coord[1] = -coord[1]

    return profile, CPs


def find_contours(image):
    """
    Calls cv2.findContours() on passed image in a way that is compatible with OpenCV 4.x, 3.x or 2.x
    versions. Passed image is a numpy.array.

    Note, cv2.findContours() will treat non-zero pixels as 1 and zero pixels as 0, so the edges detected will only
    be those on the boundary of pixels with non-zero and zero values.

    Returns a numpy array of the contours in descending arc length order.
    """
    if len(image.shape) > 2:
        raise ValueError(f"image {image} must be a single channel image")

    if CV2_VERSION >= (4, 0, 0):
        # In OpenCV 4.0, cv2.findContours() no longer returns three arguments, it reverts to the same return signature
        # as pre 3.2.0.
        contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )
    elif CV2_VERSION >= (3, 2, 0):
        # In OpenCV 3.2, cv2.findContours() does not modify the passed image and instead returns the
        # modified image as the first, of the three, return values.
        _, contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )
    else:
        contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )

    # Each contour has shape (n, 1, 2) where 'n' is the number of points. Presumably this is so each
    # point is a size 2 column vector, we don't want this so reshape it to a (n, 2)
    contours = [contour.reshape(contour.shape[0], 2) for contour in contours]

    # Sort the contours by arc length, descending order
    contours.sort(key=lambda c: cv2.arcLength(c, False), reverse=True)

    return contours


def extract_edges_CV(img):
    """
    give the image and return a list of [x.y] coordinates for the detected edges

    """
    IGNORE_EDGE_MARGIN = 1
    img = img.astype("uint8")
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    # ret, thresh = cv2.threshold(gray,threshValue,255,cv2.THRESH_BINARY)
    ret, thresh = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    contours, hierarchy = cv2.findContours(thresh, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
    # Each contour has shape (n, 1, 2) where 'n' is the number of points. Presumably this is so each
    # point is a size 2 column vector, we don't want this so reshape it to a (n, 2)
    contours = [contour.reshape(contour.shape[0], 2) for contour in contours]

    # Sort the contours by arc length, descending order
    contours.sort(key=lambda c: cv2.arcLength(c, False), reverse=True)

    # Assume that the drop is the largest contour
    # drop_profile = contours[0]
    drop_profile = contours[0]

    # Put the drop contour coordinates in order (from ?? to ??)
    # drop_profile = squish_contour(drop_profile)

    # Ignore points of the drop profile near the edges of the drop image
    width, height = img.shape[1::-1]
    if not (width < IGNORE_EDGE_MARGIN or height < IGNORE_EDGE_MARGIN):
        mask = (
            (IGNORE_EDGE_MARGIN < drop_profile[:, 0])
            & (drop_profile[:, 0] < width - IGNORE_EDGE_MARGIN)
            & (IGNORE_EDGE_MARGIN < drop_profile[:, 1])
            & (drop_profile[:, 1] < height - IGNORE_EDGE_MARGIN)
        )
        drop_profile = drop_profile[mask]

    output = []
    for coord in drop_profile:
        if list(coord) not in output:
            output.append(list(coord))
    output = np.array(output)
    return output


def ellipse_line_intersection(xc, yc, a, b, theta, x0, y0, x1, y1, display=False):
    """
    Finds the intersection between an ellipse and a line defined by two points.

    Parameters:
    xc (float): X-coordinate of the center of the ellipse.
    yc (float): Y-coordinate of the center of the ellipse.
    a (float): Length of the semi-major axis of the ellipse.
    b (float): Length of the semi-minor axis of the ellipse.
    theta (float): Angle of rotation of the ellipse in radians.
    x0 (float): X-coordinate of the first point along the line.
    y0 (float): Y-coordinate of the first point along the line.
    x1 (float): X-coordinate of the second point along the line.
    y1 (float): Y-coordinate of the second point along the line.
    display (Boolean): Set to True to show figures

    Returns:
    tuple: A tuple containing two tuples, where each inner tuple contains the X and Y coordinates of the intercept and the gradient of the ellipse at the intercept.

    """
    # Plot the ellipse and the baseline
    if display:
        ell = Ellipse(
            [xc, yc],
            2 * a,
            2 * b,
            theta,
            facecolor=(1, 0, 0, 0),
            edgecolor=(1, 0, 0, 1),
        )
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.add_patch(ell)
        ax.plot(x0, y0, "yo")
        ax.plot(x1, y1, "go")
        ax.plot()
        plt.title("Drawn ellipse with line points in green and yellow")
        plt.gca().invert_yaxis()
        plt.show()
        plt.close()

    # Calculate the slope and intercept of the line
    m = (y1 - y0) / (x1 - x0)
    c = y0 - m * x0

    # equation for ellipse in quadratic coefficient form
    A = np.sin(theta) ** 2 / b**2 + np.cos(theta) ** 2 / a**2
    B = (
        -2 * np.sin(theta) * np.cos(theta) / b**2
        + 2 * np.sin(theta) * np.cos(theta) / a**2
    )
    C = np.cos(theta) ** 2 / b**2 + np.sin(theta) ** 2 / a**2
    D = (
        -2 * xc * np.sin(theta) ** 2 / b**2
        + 2 * yc * np.sin(theta) * np.cos(theta) / b**2
        - 2 * xc * np.cos(theta) ** 2 / a**2
        - 2 * yc * np.sin(theta) * np.cos(theta) / a**2
    )
    E = (
        2 * xc * np.sin(theta) * np.cos(theta) / b**2
        - 2 * yc * np.cos(theta) ** 2 / b**2
        - 2 * xc * np.sin(theta) * np.cos(theta) / a**2
        - 2 * yc * np.sin(theta) ** 2 / a**2
    )
    F = (
        -1
        + xc**2 * np.sin(theta) ** 2 / b**2
        - 2 * xc * yc * np.sin(theta) * np.cos(theta) / b**2
        + yc**2 * np.cos(theta) ** 2 / b**2
        + xc**2 * np.cos(theta) ** 2 / a**2
        + 2 * xc * yc * np.sin(theta) * np.cos(theta) / a**2
        + yc**2 * np.sin(theta) ** 2 / a**2
    )

    # sub (mx + b) in for y, expand and simplify
    A_new = A + B * m + C * m**2
    B_new = B * c + 2 * C * m * c + D + E * m
    C_new = C * c**2 + E * c + F

    # Calculate the discriminant of the quadratic equation
    disc = B_new**2 - 4 * A_new * C_new

    # If the discriminant is negative, there are no intercepts
    if disc < 0:
        print("The line does not intersect the ellipse.")
        return None

    # If the discriminant is zero, there is one intercept
    elif disc == 0:
        x_int = -B_new / (2 * A_new)
        y_int = m * x_int + c
        x_int_rot = x_int * np.cos(theta) - y_int * np.sin(theta)
        y_int_rot = x_int * np.sin(theta) + y_int * np.cos(theta)
        grad = -(np.cos(theta) ** 2 * x_int_rot) / (np.sin(theta) ** 2 * y_int_rot)

        return ((x_int, y_int, grad),)

    # If the discriminant is positive, there are two intercepts
    else:
        x_int_1 = (-B_new + np.sqrt(disc)) / (2 * A_new)
        x_int_2 = (-B_new - np.sqrt(disc)) / (2 * A_new)
        y_int_1 = m * x_int_1 + c
        y_int_2 = m * x_int_2 + c

        grad_1 = -(
            2
            * ((-xc + x_int_1) * np.sin(theta) + (-yc + y_int_1) * np.cos(theta))
            * np.sin(theta)
            / b**2
            + 2
            * ((-xc + x_int_1) * np.cos(theta) + (-yc + y_int_1) * np.sin(theta))
            * np.cos(theta)
            / a**2
        ) / (
            2
            * ((-xc + x_int_1) * np.sin(theta) + (-yc + y_int_1) * np.cos(theta))
            * np.cos(theta)
            / b**2
            + 2
            * ((-xc + x_int_1) * np.cos(theta) + (-yc + y_int_1) * np.sin(theta))
            * np.sin(theta)
            / a**2
        )
        grad_2 = -(
            2
            * ((-xc + x_int_1) * np.sin(theta) + (-yc + y_int_2) * np.cos(theta))
            * np.sin(theta)
            / b**2
            + 2
            * ((-xc + x_int_2) * np.cos(theta) + (-yc + y_int_2) * np.sin(theta))
            * np.cos(theta)
            / a**2
        ) / (
            2
            * ((-xc + x_int_2) * np.sin(theta) + (-yc + y_int_2) * np.cos(theta))
            * np.cos(theta)
            / b**2
            + 2
            * ((-xc + x_int_2) * np.cos(theta) + (-yc + y_int_2) * np.sin(theta))
            * np.sin(theta)
            / a**2
        )

        return ((x_int_1, y_int_1, grad_1), (x_int_2, y_int_2, grad_2))


def ellipse_closest_point(xp, yp, xc, yc, a, b, th, n=1000, display=False):
    """
    xp (float): The x-coordinate of the reference point
    yp (float): The y-coordinate of the reference point
    xc (float): The x-coordinate of the ellipse's center.
    yc (float): The y-coordinate of the ellipse's center.
    a (float): The semi-major axis length of the ellipse.
    b (float): The semi-minor axis length of the ellipse.
    th (float): The rotation angle of the ellipse in degrees.
    n (int): The number of discrete points used to draw the ellipse.
    display (Boolean): Set to True to output figures and information.

    Returns:
        The distance between the reference point and the ellipse edge, and
        the coordinates of the closest point on the ellipse edge.

    """

    t = np.linspace(0, 2 * np.pi, n)

    x = (
        xc
        + a * np.cos(t) * np.cos(np.deg2rad(th))
        - b * np.sin(t) * np.sin(np.deg2rad(th))
    )
    y = (
        yc
        + a * np.cos(t) * np.sin(np.deg2rad(th))
        + b * np.sin(t) * np.cos(np.deg2rad(th))
    )

    # tRandom = np.random.randint(n)

    # xp = x[tRandom]
    # yp = y[tRandom]

    dist = np.sqrt((x - xp) ** 2 + (y - yp) ** 2)
    idx = list(dist).index(min(dist))

    # ddistdt = ((b ** 2 - a ** 2) * np.cos(t) + a * np.sin(np.deg2rad(th)) * yp - a * np.sin(np.deg2rad(th)) * yc + a * np.cos(np.deg2rad(th)) * xp - a * np.cos(np.deg2rad(th)) * xc) * np.sin(t) + ((-b * np.cos(np.deg2rad(th)) * yp) + b * np.cos(np.deg2rad(th)) * yc + b * np.sin(np.deg2rad(th)) * xp - b * np.sin(np.deg2rad(th)) * xc) * np.cos(t)
    # idx = np.where(ddistdt[1:] * ddistdt[:-1] < 0)[0]  # find zeros
    # m = (ddistdt[idx + 1] - ddistdt[idx]) / (t[idx + 1] - t[idx])  # slope
    if display:
        plt.figure(1)
        plt.plot(x, y, "-", xp, yp, "r+", x[idx], y[idx], "r+")
        plt.xlabel("x")
        plt.ylabel("y")
        plt.title("Ellipse, Point, and Zeros")

        plt.figure(2)
        plt.plot(t, dist, "m.", t[idx], dist[idx], "cx")
        plt.xlabel("t")
        plt.ylabel("Distance")
        plt.title("Distance Function")

        print(f"xp: {xp}, x[idx]: {x[idx]}")
        print(f"yp: {yp}, y[idx]: {y[idx]}")
        print("Error is: ", dist[idx])

        plt.show()
        plt.close()

    return dist[idx], [x[idx], y[idx]]


def ellipse_fit_errors(contour, h, k, a, b, theta, display):
    """
    Calculates the minimum distance between a point and the edge of a rotated and translated ellipse.

    Parameters:
        contour (array): The array of x, y coordindate points
        h (float): The x-coordinate of the ellipse's center.
        k (float): The y-coordinate of the ellipse's center.
        a (float): The semi-major axis length of the ellipse.
        b (float): The semi-minor axis length of the ellipse.
        theta (float): The rotation angle of the ellipse in degrees.
        display (boolean): Set to true to show figures.

    Returns:
        dictionary: The MAE, MSE, RMSE, and maximum error of the contour as compared against the
        fitted ellipse.
    """

    errors = []

    for point in contour:
        dist2edge, edge_point = ellipse_closest_point(
            point[0], point[1], h, k, a, b, theta
        )
        errors.append(dist2edge)

    error_measures = {}

    error_measures["MAE"] = sum([abs(error) for error in errors]) / len(errors)
    error_measures["MSE"] = sum([error**2 for error in errors]) / len(errors)
    error_measures["RMSE"] = np.sqrt(sum([error**2 for error in errors]) / len(errors))
    error_measures["Maximum error"] = max(errors)

    return error_measures


def ellipse_fit_img(img, display=False):
    """Call this function to perform the circular fit.
    For best results, peprocessing must be done before calling this function.
    """
    # begin with method specific preprocessing of img data
    start_time = time.time()

    # array of x,y coords where lines are detected
    edges_pts = extract_edges_CV(img)

    if display:
        plt.imshow(img)
        plt.plot(edges_pts[:, 0], edges_pts[:, 1])
        plt.title("drop found by hough transform")
        plt.show()
        plt.close()

    drop, CPs = prepare_hydrophobic(edges_pts, display)

    # define baseline as between the two contact points

    x = drop[:, 0]
    y = drop[:, 1]
    rise = CPs[1][1] - CPs[0][1]
    run = CPs[1][0] - CPs[0][0]
    slope = rise / run
    baseline = [(CPs[0][0], CPs[0][1]), slope]
    c = CPs[0][1] - (slope * CPs[0][0])
    baseline_x = np.linspace(1, img.shape[1], 100)
    baseline_y = slope * baseline_x + c

    # timers
    fit_preprocessing_time = time.time() - start_time
    fit_start_time = time.time()

    # Center estimates
    # x estimate is where between the lowest and highest points of the top section for a hydrophobic drop
    x_m = min(x) + (max(x) - min(x)) / 2
    # for full contour, y estimate is the halfway between max y and min y
    y_m = min(y) + ((max(y) - min(y)) / 2)

    # fit
    avec = fit_ellipse(x, y)
    (a, b), phi_deg, t = ell_parameters(avec)

    ell = Ellipse(
        t, 2 * a, 2 * b, phi_deg, facecolor=(1, 0, 0, 0), edgecolor=(1, 0, 0, 1)
    )

    if display:
        print("centre points: " + str(t))
        print("a and b: " + str(a) + ", " + str(b))
        print("phi (°): " + str(phi_deg))

        # plot
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.add_patch(ell)
        ax.scatter(x, y)
        ax.plot(baseline_x, baseline_y)
        ax.plot(CPs[0][0], CPs[0][1], "yo")
        ax.plot(CPs[1][0], CPs[1][1], "go")
        ax.plot()
        plt.title("Fitted ellipse")
        plt.imshow(img)
        plt.show()
        plt.close()

    # Find intercepts, and gradients at intercepts
    outputs = ellipse_line_intersection(
        t[0],
        t[1],
        a,
        b,
        math.radians(phi_deg),
        CPs[0][0],
        CPs[0][1],
        CPs[1][0],
        CPs[1][1],
    )
    left, right = list(sorted(outputs, key=lambda x: x[0]))
    m_left, m_right = left[2], right[2]
    intercepts = [[left[0], left[1]], [right[0], right[1]]]

    CA = []
    for output in outputs:
        first = True
        m = output[2] - slope

        if output[1] > t[1] and first == True:  # high CA angle left
            # CA.append(math.degrees(np.pi+np.arctan(m)))
            CA.append(180 - abs(math.degrees(np.arctan(m))))
        elif output[1] > t[1] and first == False:  # high CA angle right
            CA.append(abs(math.degrees(np.arctan(m))))
        elif output[1] < t[1] and first == True:  # low CA angle left
            CA.append(math.degrees(np.arctan(m)))
        elif output[1] < t[1] and first == False:  # low CA angle right
            CA.append(math.degrees(np.pi + np.arctan(m)))

        first = False

    # if CA[0]!=CA[1]:
    #    CA[1] = 180 - CA[1]

    fit_time = time.time() - fit_start_time

    # using MAE, MSE, RMSE, max_error as error measure
    errors = ellipse_fit_errors(drop, t[0], t[1], a, b, phi_deg, display)

    analysis_time = time.time() - start_time

    timings = {}
    timings["method specific preprocessing time"] = fit_preprocessing_time
    timings["fit time"] = fit_time
    timings["analysis time"] = analysis_time

    return CA, intercepts, t, (a, b), phi_deg, errors, timings


def ellipse_fit(drop, display=False):
    """Call this function to perform the circular fit.
    For best results, peprocessing must be done before calling this function.
    """
    # begin with method specific preprocessing of img data
    start_time = time.time()
    CPs = [drop[0], drop[-1]]

    # define baseline as between the two contact points

    x, y = drop[:, 0], drop[:, 1]
    rise = CPs[1][1] - CPs[0][1]
    run = CPs[1][0] - CPs[0][0]
    slope = rise / run
    baseline = [(CPs[0][0], CPs[0][1]), slope]
    c = CPs[0][1] - (slope * CPs[0][0])
    baseline_x = np.linspace(1, max(drop[:, 0]), 100)
    baseline_y = slope * baseline_x + c

    # fit
    fit_ell = EllipseModel()
    fit_ell.estimate(drop)
    xc, yc, a, b, phi_deg = fit_ell.params
    t = [xc, yc]

    ell = Ellipse(
        t,
        2 * a,
        2 * b,
        phi_deg,
        facecolor=(1, 0, 0, 0),
        edgecolor=(1, 0, 0, 1),
        label="Fitted ellipse",
    )

    if display:
        print("drop contour: ", drop)
        print("centre points: " + str(t))
        print("a and b: " + str(a) + ", " + str(b))
        print("phi (°): " + str(phi_deg))

        if 0:  # plot
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.add_patch(ell)
            ax.scatter(x, y, label="contour")
            ax.plot(baseline_x, baseline_y, label="baseline")
            ax.plot(CPs[0][0], CPs[0][1], "yo", label="left contact point")
            ax.plot(CPs[1][0], CPs[1][1], "go", label="right contact point")
            ax.plot()
            ax.legend()
            plt.title("Fitted ellipse")
            plt.axis("equal")
            plt.show()
            plt.close()

    # Find intercepts, and gradients at intercepts
    outputs = ellipse_line_intersection(
        t[0],
        t[1],
        a,
        b,
        math.radians(phi_deg),
        CPs[0][0],
        CPs[0][1],
        CPs[1][0],
        CPs[1][1],
    )
    left, right = list(sorted(outputs, key=lambda x: x[0]))
    m_left, m_right = left[2], right[2]
    intercepts = [[left[0], left[1]], [right[0], right[1]]]

    CA = []
    for output in outputs:
        first = True
        m = output[2] - slope

        if output[1] > t[1] and first == True:  # high CA angle left
            # CA.append(math.degrees(np.pi+np.arctan(m)))
            CA.append(180 - abs(math.degrees(np.arctan(m))))
        elif output[1] > t[1] and first == False:  # high CA angle right
            CA.append(abs(math.degrees(np.arctan(m))))
        elif output[1] < t[1] and first == True:  # low CA angle left
            CA.append(math.degrees(np.arctan(m)))
        elif output[1] < t[1] and first == False:  # low CA angle right
            CA.append(math.degrees(np.pi + np.arctan(m)))

        first = False

    # if CA[0]!=CA[1]:
    #    CA[1] = 180 - CA[1]

    fit_time = time.time() - start_time

    # using MAE, MSE, RMSE, max_error as error measure
    errors = ellipse_fit_errors(drop, t[0], t[1], a, b, phi_deg, display)

    analysis_time = time.time() - start_time

    timings = {}
    timings["fit time"] = fit_time
    timings["analysis time"] = analysis_time

    return CA, intercepts, t, (a, b), phi_deg, errors, timings


if 0:
    from utils.misc import resource_path
    IMG_PATH = resource_path("../RICOphobic_cropped.png")
    img = cv2.imread(IMG_PATH)

    angles, intercepts, center, (a, b), theta, errors, timings = ellipse_fit_img(
        img, display=True
    )
    print("angles: ", angles)
    print("intercepts: ", intercepts)
    print("center coordinates: ", center)
    print("a: ", a)
    print("b: ", b)
    print("angle of ellipse rotation: ", theta)
    print("errors: ", errors)
    print("timings: ", timings)



#!/usr/bin/env python
# coding=utf-8

from math import pi
from numpy import sin, cos

# minimise calls to sin() and cos()
# defines the Young--Laplace system of differential equations to be solved


def ylderiv(x_vec, t, bond_number):
    x, y, phi, x_Bond, y_Bond, phi_Bond = x_vec
    x_s = cos(phi)
    y_s = sin(phi)
    phi_s = 2 + bond_number * y - y_s / x  # DS 16/09 2- made 2+ for contact angle
    x_Bond_s = -y_s * phi_Bond
    y_Bond_s = x_s * phi_Bond
    phi_Bond_s = y_s * x_Bond / (x * x) - x_s * phi_Bond / x - y - bond_number * y_Bond
    return [x_s, y_s, phi_s, x_Bond_s, y_Bond_s, phi_Bond_s]


# defines the Young--Laplace system of differential equations to be solved


def dataderiv(x_vec, t, bond_number):
    x, y, phi, vol, sur = x_vec
    x_s = cos(phi)
    y_s = sin(phi)
    phi_s = 2 - bond_number * y - sin(phi) / x
    vol_s = pi * x**2 * y_s
    sur_s = 2 * pi * x
    return [x_s, y_s, phi_s, vol_s, sur_s]


# # defines the Young--Laplace system of differential equations to be solved
# def ylderiv(x_vec, t):
#     x, y, phi, x_Bond, y_Bond, phi_Bond = x_vec
#     x_s = cos(phi)
#     y_s = sin(phi)
#     phi_s = 2 - bond_number * y - sin(phi)/x
#     x_Bond_s = -sin(phi)*phi_Bond
#     y_Bond_s = cos(phi)*phi_Bond
#     phi_Bond_s = sin(phi) * x_Bond / (x*x) - cos(phi) * phi_Bond / x - y - bond_number * y_Bond
#     return [x_s, y_s, phi_s, x_Bond_s, y_Bond_s, phi_Bond_s]



from utils.geometry import Rect2

from typing import Sequence, Tuple, NamedTuple, Optional
from enum import IntEnum, auto
from scipy.ndimage import gaussian_filter
import math
import numpy as np
import scipy.signal
import scipy.optimize

try:
    from modules.ift.hough import hough
except ImportError:
    raise RuntimeError(
        "❗ Failed to load native Cython module 'hough'.\n"
        "This is likely due to a missing or incompatible build for your architecture.\n"
        "Please run: `python setup.py build_ext --inplace`\n"
        "Refer to the README section: 'Troubleshooting: Architecture Mismatch (macOS)'"
    )


__all__ = (
    "NeedleFitResult",
    "needle_fit",
)

DELTA_TOL = None
GRADIENT_TOL = 1.0e-8
OBJECTIVE_TOL = 1.0e-8

ANGLE_STEPS = 200
DIST_STEPS = 64

# def hough(data: np.ndarray, diagonal: float) -> np.ndarray:
#     votes = np.zeros((ANGLE_STEPS, DIST_STEPS + 2), dtype=np.int32)
#     thetas = np.arange(ANGLE_STEPS) / ANGLE_STEPS * np.pi
#     for i in range(data.shape[1]):
#         x = data[0, i]
#         y = data[1, i]
#         rhos = x * np.sin(thetas) - y * np.cos(thetas)
#         bins = np.round((rhos + 0.5 * diagonal) / diagonal * (DIST_STEPS - 1)).astype(int) + 1
#         for theta_i, rho_i in enumerate(bins):
#             votes[theta_i, rho_i] += 1
#     return votes


class NeedleFitResult(NamedTuple):
    rotation: float
    rho: float
    radius: float

    objective: float
    residuals: np.ndarray

    lmask: np.ndarray


def needle_fit(
    data: Tuple[np.ndarray, np.ndarray], verbose: bool = False
) -> Optional[NeedleFitResult]:
    if data.shape[1] == 0:
        return None

    model = NeedleModel(data)

    def fun(params: Sequence[float], model: NeedleModel) -> np.ndarray:
        model.set_params(params)
        residuals = model.residuals.copy()
        return residuals

    def jac(params: Sequence[float], model: NeedleModel) -> np.ndarray:
        model.set_params(params)
        jac = model.jac.copy()
        return jac

    try:
        optimize_result = scipy.optimize.least_squares(
            fun,
            needle_guess(data),
            jac,
            args=(model,),
            x_scale="jac",
            method="trf",
            loss="arctan",
            f_scale=2.0,
            ftol=OBJECTIVE_TOL,
            xtol=DELTA_TOL,
            gtol=GRADIENT_TOL,
            max_nfev=50,
            verbose=2 if verbose else 0,
        )
    except ValueError:
        return None

    # Update model parameters to final result.
    model.set_params(optimize_result.x)

    result = NeedleFitResult(
        rotation=model.params[NeedleParam.ROTATION],
        rho=model.params[NeedleParam.RHO],
        radius=math.fabs(model.params[NeedleParam.RADIUS]),
        objective=(model.residuals**2).sum() / model.dof,
        residuals=model.residuals,
        lmask=model.lmask,
    )

    return result


def needle_guess(data: np.ndarray) -> Sequence[float]:
    params = np.empty(len(NeedleParam))
    data = data.astype(float)

    extents = Rect2(data.min(axis=1), data.max(axis=1))
    diagonal = int(math.ceil((extents.w**2 + extents.h**2) ** 0.5))
    data -= np.reshape(extents.center, (2, 1))
    votes = hough.hough(data, diagonal)

    needles = np.zeros(shape=(votes.shape[0], 3))
    for i in range(votes.shape[0]):
        peaks, props = scipy.signal.find_peaks(votes[i], prominence=0)
        if len(peaks) < 2:
            continue
        ix = np.argsort(props["prominences"])[::-1]
        peak1_i, peak2_i = peaks[ix[:2]]
        prom1, prom2 = props["prominences"][ix[:2]]
        if prom2 < prom1 / 2:
            continue

        peak1 = ((peak1_i - 1) / (len(votes[i]) - 3) - 0.5) * diagonal
        peak2 = ((peak2_i - 1) / (len(votes[i]) - 3) - 0.5) * diagonal

        needles[i][0] = (peak1 + peak2) / 2
        needles[i][1] = math.fabs(peak1 - peak2) / 2
        needles[i][2] = prom1 + prom2

    scores = gaussian_filter(needles[:, 2], sigma=10, mode="wrap")
    needle_i = scores.argmax()

    theta = -np.pi / 2 + (needle_i / len(needles)) * np.pi
    rho, radius = needles[needle_i][:2]

    rho_offset = np.cos(theta) * extents.xc + np.sin(theta) * extents.yc
    rho += rho_offset

    params[NeedleParam.ROTATION] = theta
    params[NeedleParam.RHO] = rho
    params[NeedleParam.RADIUS] = radius

    return params


class NeedleModel:
    def __init__(self, data: Tuple[np.ndarray, np.ndarray]) -> None:
        self.data = np.copy(data)
        self.data.flags.writeable = False

        self._params = np.empty(len(NeedleParam))
        self._params_set = False
        self._residuals = np.empty(shape=(self.data.shape[1],))
        self._jac = np.empty(shape=(self.data.shape[1], len(self._params)))
        self._lmask = np.empty(shape=(self.data.shape[1],), dtype=bool)

    def set_params(self, params: Sequence[float]) -> None:
        if self._params_set and (self._params == params).all():
            return

        w = params[NeedleParam.ROTATION]
        rho = params[NeedleParam.RHO]
        radius = params[NeedleParam.RADIUS]

        residuals = self._residuals
        de_dw = self._jac[:, NeedleParam.ROTATION]
        de_drho = self._jac[:, NeedleParam.RHO]
        de_dR = self._jac[:, NeedleParam.RADIUS]
        lmask = self._lmask

        Q = np.array([[np.cos(w), -np.sin(w)], [np.sin(w), np.cos(w)]])

        data_x, data_y = self.data
        data_r, data_z = Q.T @ (data_x, data_y) - [[rho], [0]]

        e = np.abs(data_r) - radius

        lmask[:] = data_r < 0
        rmask = ~lmask

        residuals[:] = e
        de_dw[rmask] = data_z[rmask]
        de_dw[lmask] = -data_z[lmask]
        de_dR[:] = -1
        de_drho[rmask] = -1
        de_drho[lmask] = 1

        self._params[:] = params

    @property
    def params(self) -> Sequence[int]:
        params = self._params[:]
        params.flags.writeable = False
        return params

    @property
    def dof(self) -> int:
        return self.data.shape[1] - len(self.params) + 1

    @property
    def jac(self) -> np.ndarray:
        jac = self._jac[:]
        jac.flags.writeable = False
        return jac

    @property
    def residuals(self) -> np.ndarray:
        residuals = self._residuals[:]
        residuals.flags.writeable = False
        return residuals

    @property
    def lmask(self) -> np.ndarray:
        lmask = self._lmask[:]
        lmask.flags.writeable = False
        return lmask


class NeedleParam(IntEnum):
    ROTATION = 0
    RHO = auto()
    RADIUS = auto()
